import numpy as np


def test_placement(case_data):
    """
    Test placement methods.

    This tests both :meth:`.Vertex.place_building_block` and
    :meth:`.Vertex.map_functional_groups_to_edges`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the vertex to test, the building block which
        it  places, and the correct position and alignment of the
        building block after placement.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_placement(
        # Use a clone to ensure that the clone() method is implemented
        # correctly.
        vertex=case_data.vertex.clone(),
        edges=case_data.edges,
        building_block=case_data.building_block,
        position=case_data.position,
        alignment_tests=case_data.alignment_tests,
        functional_group_edges=case_data.functional_group_edges,
        position_ids=case_data.position_ids,
    )


def _test_placement(
    vertex,
    edges,
    building_block,
    position,
    alignment_tests,
    functional_group_edges,
    position_ids,
):
    """
    Test placement methods.

    This tests both :meth:`.Vertex.place_building_block` and
    :meth:`.Vertex.map_functional_groups_to_edges`.

    Parameters
    ----------
    vertex : :class:`.Vertex`
        The vertex to test.

    edges : :class:`tuple` of :class:`.Edge`
        The edges connected to `vertex`.

    building_block : :class:`.BuildingBlock`
        The building block which is placed by `vertex`.

    position : :class:`numpy.ndarray`
        The correct centroid of atoms in `position_ids` after
        placement.

    alignment_tests : :class:`dict`
        Maps a :class:`callable` to a :class:`numpy.ndarray`.
        Each :class:`callable` takes a singe parameter,
        `building_block`, and returns a :class:`numpy.ndarray`.

        Each :class:`callable` in `alignment_tests` represents a
        test, and the array which it maps to is the correct result
        for that test. If the array returned by the :class:`callable`
        does not match the array it is mapped to, the test will fail.

        The point of these tests is to make sure that `building_block`
        is aligned correctly after placement. Therefore, alignment
        tests should return some vector which depends on the specific
        orientation of the building block.

    functional_group_edges : :class:`dict`
        The correct mapping of functional group id to edge id.

    position_ids : :class:`tuple` of :class:`int`
        The ids of atoms which are used to determine if the building
        block was positioned correctly.

    Returns
    -------
    None : :class:`NoneType`

    """

    position_matrix = vertex.place_building_block(
        building_block=building_block,
        edges=edges,
    )
    building_block = building_block.with_position_matrix(
        position_matrix=position_matrix,
    )
    assert np.allclose(
        a=building_block.get_centroid(position_ids),
        b=position,
        atol=1e-14,
    )
    for test, result in alignment_tests.items():
        # Do the assert via a function call here so that pytest prints
        # the values being compared when an error occurs.
        assert_equal_vectors(test(building_block), result)

    functional_group_edges_ = vertex.map_functional_groups_to_edges(
        building_block=building_block,
        edges=edges,
    )
    assert functional_group_edges_ == functional_group_edges


def assert_equal_vectors(vector1, vector2):
    assert np.all(np.equal(vector2, vector1))



class CaseData:
    """
    A test case.

    Attributes
    ----------
    edge : :class:`.Edge`
        The edge to test.

    id : :class:`int`
        The correct id of the edge.

    vertex1_id : :class:`int`
        The correct id of the first vertex held by the edge.

    vertex2_id : :class:`int`
        The correct id of the second vertex held by the edge.

    periodicity : :class:`tuple` of :class:`int`
        The correct periodicity of :attr:`.edge`.

    is_periodic : :class:`bool`
        The truth about :attr:`.edge` being periodic.

    """

    def __init__(
        self,
        edge,
        id,
        vertex1_id,
        vertex2_id,
        periodicity,
        is_periodic,
    ):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        edge : :class:`.Edge`
            The edge to test.

        id : :class:`int`
            The correct id of the edge.

        vertex1_id : :class:`int`
            The correct id of the first vertex held by the edge.

        vertex2_id : :class:`int`
            The correct id of the second vertex held by the edge.

        periodicity : :class:`tuple` of :class:`int`
            The correct periodicity of `edge`.

        is_periodic : :class:`bool`
            The truth about `edge` being periodic.

        """

        self.edge = edge
        self.id = id
        self.vertex1_id = vertex1_id
        self.vertex2_id = vertex2_id
        self.periodicity = periodicity
        self.is_periodic = is_periodic



import pytest

import stk

from ....case_data import CaseData
from ...building_blocks import get_fe_atom, get_mo_1


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.Octahedral(
                    metals={get_fe_atom(): 0},
                    ligands={get_mo_1(): (0, 1, 2, 3, 4, 5)},
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9,
                            }
                        )
                    ),
                )
            ),
            smiles=(
                "[H]C1=C([H])[N](->[Fe+2](<-[N]2=C([H])C3=C(C([H])=C2[H])C(["
                "H])([H])C([H])([H])C([H])([H])C([H])([H])C3([H])[H])(<-[N]2"
                "=C([H])C3=C(C([H])=C2[H])C([H])([H])C([H])([H])C([H])([H])C"
                "([H])([H])C3([H])[H])(<-[N]2=C([H])C3=C(C([H])=C2[H])C([H])"
                "([H])C([H])([H])C([H])([H])C([H])([H])C3([H])[H])(<-[N]2=C("
                "[H])C3=C(C([H])=C2[H])C([H])([H])C([H])([H])C([H])([H])C([H"
                "])([H])C3([H])[H])<-[N]2=C([H])C3=C(C([H])=C2[H])C([H])([H]"
                ")C([H])([H])C([H])([H])C([H])([H])C3([H])[H])=C([H])C2=C1C("
                "[H])([H])C([H])([H])C([H])([H])C([H])([H])C2([H])[H]"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.Octahedral(
                    metals=get_fe_atom(),
                    ligands=get_mo_1(),
                    reaction_factory=stk.DativeReactionFactory(
                        reaction_factory=stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9,
                            }
                        )
                    ),
                )
            ),
            smiles=(
                "[H]C1=C([H])[N](->[Fe+2](<-[N]2=C([H])C3=C(C([H])=C2[H])C(["
                "H])([H])C([H])([H])C([H])([H])C([H])([H])C3([H])[H])(<-[N]2"
                "=C([H])C3=C(C([H])=C2[H])C([H])([H])C([H])([H])C([H])([H])C"
                "([H])([H])C3([H])[H])(<-[N]2=C([H])C3=C(C([H])=C2[H])C([H])"
                "([H])C([H])([H])C([H])([H])C([H])([H])C3([H])[H])(<-[N]2=C("
                "[H])C3=C(C([H])=C2[H])C([H])([H])C([H])([H])C([H])([H])C([H"
                "])([H])C3([H])[H])<-[N]2=C([H])C3=C(C([H])=C2[H])C([H])([H]"
                ")C([H])([H])C([H])([H])C([H])([H])C3([H])[H])=C([H])C2=C1C("
                "[H])([H])C([H])([H])C([H])([H])C([H])([H])C2([H])[H]"
            ),
            name=name,
        ),
    ),
)
def metal_complex_octahedral(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



"""
Two-Two Reaction
================

"""

import itertools as it

from scipy.spatial.distance import euclidean

from stk._internal.bond import Bond

from .reaction.reaction import Reaction


class TwoTwoReaction(Reaction):
    """
    A reaction between two functional groups, each with 2 bonder atoms.

    The reaction creates the two shortest possible bonds between the
    *bonder* atoms of the two functional groups, and deletes any
    *deleter* atoms.

    """

    def __init__(
        self,
        construction_state,
        functional_group1,
        functional_group2,
        bond_order,
        periodicity,
    ):
        """
        Initialize a :class:`.TwoTwoReaction` instance.

        Parameters
        ----------
        construction_state : :class:`.ConstructionState`
            The construction state of the molecule being constructed.

        functional_group1 : :class:`.GenericFunctionalGroup`
            The first functional group in the reaction.

        functional_group2 : :class:`.GenericFunctionalGroup`
            The second functional group in the reaction.

        bond_order : :class:`int`
            The bond order of the bonds created by the reaction.

        periodicity : :class:`tuple` of :class:`int`
            The periodicity of the bonds created by the reaction.

        """

        self._position_matrix = construction_state.get_position_matrix()
        self._functional_group1 = functional_group1
        self._functional_group2 = functional_group2
        self._bond_order = bond_order
        self._periodicity = periodicity

    def _get_new_atoms(self):
        return
        yield

    def _get_new_bonds(self):
        for bonder1, bonder2 in self._get_bonder_pairs():
            yield Bond(
                atom1=bonder1,
                atom2=bonder2,
                order=self._bond_order,
                periodicity=self._periodicity,
            )

    def _get_bonder_pairs(self):
        pairs = it.product(
            self._functional_group1.get_bonders(),
            self._functional_group2.get_bonders(),
        )
        sorted_pairs = sorted(pairs, key=self._pair_distance)
        bonded = set()
        for bonder1, bonder2 in sorted_pairs:
            if (
                bonder1.get_id() not in bonded
                and bonder2.get_id() not in bonded
            ):
                bonded.add(bonder1.get_id())
                bonded.add(bonder2.get_id())
                yield bonder1, bonder2

    def _pair_distance(self, bonders):
        bonder1, bonder2 = bonders
        return euclidean(
            self._position_matrix[bonder1.get_id()],
            self._position_matrix[bonder2.get_id()],
        )

    def _get_deleted_atoms(self):
        yield from self._functional_group1.get_deleters()
        yield from self._functional_group2.get_deleters()

    def _get_deleted_bonds(self):
        return
        yield



import itertools as it


def atom_id(atom):
    return atom.get_id()


def are_same_id_sequences(ids1, ids2):
    for id1, id2 in it.zip_longest(ids1, ids2):
        assert id1 == id2


def are_clone_sequences(atoms1, atoms2):
    """
    Test if `atoms1` and `atoms2` are clones of each other.

    """

    for a1, a2 in it.zip_longest(atoms1, atoms2):
        assert a1 is not a2
        assert a1.get_id() == a2.get_id()
        assert a1.get_charge() == a2.get_charge()
        assert a1.__class__ is a2.__class__



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def carboxylic_acid(get_atom_ids):
    a, b, c, d, e = get_atom_ids(5)
    return _carboxylic_acid(
        carbon=stk.C(a),
        oxygen1=stk.O(b),
        oxygen2=stk.O(c),
        hydrogen=stk.H(d),
        atom=stk.C(e),
    )


def _carboxylic_acid(carbon, oxygen1, oxygen2, hydrogen, atom):
    bonders = (carbon,)
    deleters = (oxygen2, hydrogen)
    return GenericCaseData(
        functional_group=stk.CarboxylicAcid(
            carbon=carbon,
            oxygen1=oxygen1,
            oxygen2=oxygen2,
            hydrogen=hydrogen,
            atom=atom,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(carbon, oxygen1, oxygen2, hydrogen, atom),
        bonders=bonders,
        deleters=deleters,
    )



import typing
from collections.abc import Callable, Iterable
from typing import Any

import numpy as np

from .fitness_normalizer import FitnessNormalizer

T = typing.TypeVar("T")


class Multiply(FitnessNormalizer[T]):
    """
    Multiplies the fitness values by some coefficient.

    Examples:

        *Multiplying Fitness Values by a Set of Coefficients*

        In this example, assume that each fitness value consists of a
        :class:`tuple` of numbers, each representing a different property
        of the molecule, and each contributing to the final fitness value.
        The properties can be anything, such as  energy, number of atoms
        or diameter.

        If your final fitness value depends on the combination of these
        properties, you will probably first want to scale them with
        :class:`.DivideByMean`. Once this is done, you may want to
        multiply each property by some coefficient, which reflects its
        relative importance to the final fitness value. For example
        if you multiply the value of one property by ``1`` and another
        by ``2``, the second will contribute twice as much to the
        final fitness value, assuming that you get the final fitness value
        by using the :class:`.Sum` normalizer after :class:`.Multiply`.

        Giving a concrete example

        .. testcode:: multiplying-fitness-values-by-a-set-of-coefficients

            import stk
            import numpy as np

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record: (1, 1, 1),
            }
            normalizer = stk.Multiply((1, 2, 3))
            normalized_fitness_values = normalizer.normalize(fitness_values)
            assert np.all(np.equal(
                normalized_fitness_values[record],
                (1, 2, 3),
            ))

        *Selectively Normalizing Fitness Values*

        Sometimes, you only want to normalize some members of a population,
        for example if some do not have an assigned fitness value,
        because the fitness calculation failed for whatever reason.
        You can use the `filter` parameter to exclude records from the
        normalization

        .. testcode:: selectively-normalizing-fitness-values

            import stk
            import numpy as np

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record1: (1, 1, 1),
                record2: None,
            }
            normalizer = stk.Multiply(
                coefficient=(1, 2, 3),
                # Only normalize values which are not None.
                filter=lambda fitness_values, record:
                    fitness_values[record] is not None,
            )
            normalized_fitness_values = normalizer.normalize(fitness_values)
            assert np.all(np.equal(
                normalized_fitness_values[record1],
                (1, 2, 3),
            ))
            assert normalized_fitness_values[record2] is None
    """

    def __init__(
        self,
        coefficient: float | Iterable[float],
        filter: Callable[[dict[T, Any], T], bool] = lambda fitness_values,
        record: True,
    ) -> None:
        """
        Parameters:
            coefficient (float | list[float]):
                The coefficients each fitness value is multiplied by. Can
                be a single number or multiple numbers, depending on the
                form of the fitness value.

            filter:
                A function which returns ``True`` or ``False``. Only
                molecules which return ``True`` will have fitness values
                normalized. By default, all molecules will have fitness
                values normalized.
                The instance passed to the `fitness_values` argument of
                :meth:`.normalize` is passed as the first argument, while
                the second argument will be passed every
                :class:`.MoleculeRecord` in it, one at a time.
        """
        if not isinstance(coefficient, int | float):
            coefficient = tuple(coefficient)
        self._coefficient = coefficient
        self._filter = filter

    def normalize(self, fitness_values: dict[T, Any]) -> dict[T, Any]:
        return {
            record: (
                np.multiply(self._coefficient, fitness_value)
                if self._filter(fitness_values, record)
                else fitness_value
            )
            for record, fitness_value in fitness_values.items()
        }



"""
Ring Amine Reaction
===================

"""

import numpy as np

from stk._internal.bond import Bond
from stk._internal.elements import C, H

from .reaction.new_atom import NewAtom
from .reaction.reaction import Reaction


class RingAmineReaction(Reaction):
    """
    Performs a reaction between two :class:`.RingAmine` groups.

    The reaction creates a new ring, merging the two functional groups.

    """

    def __init__(self, position_matrix, ring_amine1, ring_amine2):
        """
        Initialize a :class:`.RingAmineReaction` instance.

        Parameters
        ----------
        position_matrix : :class:`numpy.ndarray`
            The position matrix of the molecule being constructed.

        ring_amine1 : :class:`.RingAmine`
            The first functional group in the reaction.

        ring_amine2 : :class:`.RingAmine`
            The second functional group in the reaction.

        """

        self._position_matrix = np.array(position_matrix)
        self._ring_amine1 = ring_amine1.clone()
        self._ring_amine2 = ring_amine2.clone()

    def _get_new_atoms(self):
        n1_coord = self._get_position(self._ring_amine1.get_nitrogen())
        n2_coord = self._get_position(self._ring_amine2.get_nitrogen())
        c1_coord = self._get_position(self._ring_amine1.get_carbon1())
        c2_coord = self._get_position(self._ring_amine2.get_carbon1())

        n_joiner = C(-1)
        n_joiner_coord = (n1_coord + n2_coord) / 2
        yield NewAtom(n_joiner, n_joiner_coord)

        nh1 = H(-2)
        nh1_coord = n_joiner_coord + [0, 0, 1]
        yield NewAtom(nh1, nh1_coord)

        nh2 = H(-3)
        nh2_coord = n_joiner_coord + [0, 0, -1]
        yield NewAtom(nh2, nh2_coord)

        nc_joiner1 = C(-4)
        nc_joiner1_coord = (c1_coord + n2_coord) / 2
        yield NewAtom(nc_joiner1, nc_joiner1_coord)

        nc1h1 = H(-5)
        nc1h1_coord = nc_joiner1_coord + [0, 0, 1]
        yield NewAtom(nc1h1, nc1h1_coord)

        nc1h2 = H(-6)
        nc1h2_coord = nc_joiner1_coord + [0, 0, -1]
        yield NewAtom(nc1h2, nc1h2_coord)

        nc_joiner2 = C(-7)
        nc_joiner2_coord = (c2_coord + n1_coord) / 2
        yield NewAtom(nc_joiner2, nc_joiner2_coord)

        nc2h1 = H(-8)
        nc2h1_coord = nc_joiner2_coord + [0, 0, 1]
        yield NewAtom(nc2h1, nc2h1_coord)

        nc2h2 = H(-9)
        nc2h2_coord = nc_joiner2_coord + [0, 0, -1]
        yield NewAtom(nc2h2, nc2h2_coord)

    def _get_new_bonds(self):
        n1 = self._ring_amine1.get_nitrogen()
        n2 = self._ring_amine2.get_nitrogen()
        c1 = self._ring_amine1.get_carbon2()
        c2 = self._ring_amine2.get_carbon2()
        n_joiner = C(-1)
        nh1 = H(-2)
        nh2 = H(-3)
        nc_joiner1 = C(-4)
        nc1h1 = H(-5)
        nc1h2 = H(-6)
        nc_joiner2 = C(-7)
        nc2h1 = C(-8)
        nc2h2 = C(-9)

        yield (Bond(n1, n_joiner, 1),)
        yield (Bond(n_joiner, n2, 1, self._periodicity),)
        yield (Bond(n_joiner, nh1, 1),)
        yield (Bond(n_joiner, nh2, 1),)
        yield (Bond(c1, nc_joiner1, 1),)
        yield (Bond(nc_joiner1, n2, 1, self._periodicity),)
        yield (Bond(nc_joiner1, nc1h1, 1),)
        yield (Bond(nc_joiner1, nc1h2, 1),)
        yield (Bond(nc_joiner2, c2, 1, self._periodicity),)
        yield (Bond(n1, nc_joiner2, 1),)
        yield (Bond(nc_joiner2, nc2h1, 1),)
        yield Bond(nc_joiner2, nc2h2, 1)

    def _get_deleted_atoms(self):
        yield self._ring_amine1.get_hydrogen3()
        yield self._ring_amine2.get_hydrogen3()
        yield self._ring_amine1.get_hydrogen1()
        yield self._ring_amine2.get_hydrogen1()
        yield self._ring_amine1.get_hydrogen2()
        yield self._ring_amine2.get_hydrogen2()

    def _get_deleted_bonds(self):
        return
        yield



def test_get_num_bonders(generic_case_data):
    """
    Test :meth:`.GenericFunctionalGroup.get_num_bonders`.

    Parameters
    ----------
    generic_case_data : :class:`.GenericCaseData`
        The test case. Holds the functional group to test and the
        correct number of bonder atoms.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_num_bonders(
        functional_group=generic_case_data.functional_group,
        num_bonders=len(generic_case_data.bonders),
    )


def _test_get_num_bonders(functional_group, num_bonders):
    """
    Test :meth:`.GenericFunctionalGroup.get_num_bonders`.

    Parameters
    ----------
    functional_group : :class:`.GenericFunctionalGroup`
        The functional group to test.

    num_bonders : :class:`int`
        The correct number of bonder atoms.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert functional_group.get_num_bonders() == num_bonders



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def thiol(get_atom_ids):
    a, b, c = get_atom_ids(3)
    return _thiol(stk.S(a), stk.H(b), stk.C(c))


def _thiol(sulfur, hydrogen, atom):
    bonders = ()
    deleters = ()
    return GenericCaseData(
        functional_group=stk.Thiol(
            sulfur=sulfur,
            hydrogen=hydrogen,
            atom=atom,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(sulfur, hydrogen, atom),
        bonders=bonders,
        deleters=deleters,
    )



from stk._internal.key_makers.molecule import MoleculeKeyMaker
from stk._internal.key_makers.utilities import get_inchi_key
from stk._internal.molecule import Molecule


class InchiKey(MoleculeKeyMaker):
    """
    Used to get the InChIKey of molecules.

    Examples:

        *Adding InChIKey to a Molecule's JSON*

        You want to use the InChIKey as part of a JSON representation
        of a molecule

        .. testcode:: adding-inchikey-to-a-molecules-json

            import stk

            jsonizer = stk.MoleculeJsonizer(
                key_makers=(stk.InchiKey(), ),
            )
            # Get the JSON representation, including an InChIKey.
            json = jsonizer.to_json(stk.BuildingBlock('NCCN'))

        .. testcode:: adding-inchikey-to-a-molecules-json
            :hide:

            assert (
                json['molecule']['InChIKey']
                == 'PIICEJLVQHRZGT-UHFFFAOYSA-N'
            )
            assert (
                json['molecule']['InChIKey']
                == 'PIICEJLVQHRZGT-UHFFFAOYSA-N'
            )
    """

    def __init__(self) -> None:
        return

    def get_key_name(self) -> str:
        return "InChIKey"

    def get_key(self, molecule: Molecule) -> str:
        return get_inchi_key(molecule)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return "InchiKey()"



import os
import pathlib
import typing
from collections.abc import Iterable, Iterator

import atomlite
import numpy as np
import rdkit.Chem.AllChem as rdkit
from scipy.spatial.distance import euclidean

from stk._internal.atom import Atom
from stk._internal.bond import Bond
from stk._internal.utilities.molecule import (
    get_bond_atom_ids,
    sort_bond_atoms_by_id,
)
from stk._internal.utilities.updaters import mae, mdl_mol, pdb, turbomole, xyz
from stk._internal.utilities.utilities import (
    rotation_matrix,
    rotation_matrix_arbitrary_axis,
    vector_angle,
)
from stk._internal.utilities.writers.mdl_mol import _write_mdl_mol_file
from stk._internal.utilities.writers.pdb import _write_pdb_file
from stk._internal.utilities.writers.xyz import _write_xyz_file


class Molecule:
    """
    An abstract base class for molecules.

    Notes:

        You might notice that some of the methods of this abstract base
        class are implemented. This is purely for convenience when
        implementing subclasses. The implemented public methods are
        simply default implementations, which can be safely ignored
        or overridden, when implementing subclasses. Any private
        methods are implementation details of these default
        implementations.

    Examples:

        *Aligning a Molecule with a Vector*

        You want to rotate a molecule, such that it is aligned along
        with a specific direction.

        .. testcode:: aligning-a-molecule-with-a-vector

            import stk
            import numpy as np

            molecule1 = stk.BuildingBlock('CCCCC')
            # Align molecule1 along x-axis.
            molecule1 = molecule1.with_rotation_between_vectors(
                start=molecule1.get_direction(),
                target=np.array([1., 0., 0.]),
                origin=molecule1.get_centroid(),
            )

            molecule2 = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCCBr',
                            functional_groups=(stk.BromoFactory(), ),
                        ),
                    ),
                    repeating_unit='A',
                    num_repeating_units=15,
                ),
            )
            # Align molecule2 along the [1, 4, -3] vector.
            molecule2 = molecule2.with_rotation_between_vectors(
                start=molecule2.get_direction(),
                target=np.array([1., 4., -3.]),
                origin=molecule2.get_centroid(),
            )

        *Aligning a Molecule along a Plane*

        You want to place the benzene flat along the xy plane.

        .. testcode:: aligning-a-molecule-along-a-plane

            import stk
            import numpy as np

            benzene = stk.BuildingBlock('c1ccccc1')
            benzene = benzene.with_rotation_between_vectors(
                start=benzene.get_plane_normal(),
                target=np.array([0., 0., 1.]),
                origin=benzene.get_centroid(),
            )

    """

    def __init__(
        self,
        atoms: Iterable[Atom],
        bonds: Iterable[Bond],
        position_matrix: np.ndarray,
    ) -> None:
        """
        Initialize a :class:`.Molecule`.

        Parameters:

            atoms (list[Atom]):
                The atoms which compose the molecule.

            bonds (list[Bond]):
                The bonds of the molecule.

            position_matrix:
                A ``(n, 3)`` matrix holding the position of every atom
                in the :class:`.Molecule`.

        """

        self._atoms = tuple(atoms)
        self._bonds = tuple(bonds)
        # Take the transpose because it will make some matrix
        # multiplications faster.
        self._position_matrix = np.array(
            position_matrix.T,
            dtype=np.float64,
        )

    def _with_displacement(self, displacement: np.ndarray) -> typing.Self:
        """
        Modify molecule.

        """

        self._position_matrix = (self._position_matrix.T + displacement).T
        return self

    def with_displacement(self, displacement: np.ndarray) -> typing.Self:
        """
        Return a displaced clone.

        Parameters:

            displacement:
                The displacement vector to be applied.

        Returns:

            Molecule: A displaced clone.

        """

        return self.clone()._with_displacement(displacement)

    def _with_rotation_about_axis(
        self,
        angle: float,
        axis: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Modify molecule.

        """

        # Set the origin of the rotation to "origin".
        self._with_displacement(-origin)
        rot_mat = rotation_matrix_arbitrary_axis(angle, axis)

        # Apply the rotation matrix on the position matrix, to get the
        # new position matrix.
        self._position_matrix = rot_mat @ self._position_matrix

        # Return the centroid of the molecule to the original position.
        self._with_displacement(origin)
        return self

    def with_rotation_about_axis(
        self,
        angle: float,
        axis: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Return a rotated clone.

        The clone is rotated by `angle` about `axis` on the
        `origin`.

        Parameters:
            angle:
                The size of the rotation in radians.
            axis:
                The axis about which the rotation happens. Must have
                unit magnitude.
            origin:
                The origin about which the rotation happens.
        Returns:
            Molecule: A rotated clone.

        """
        return self.clone()._with_rotation_about_axis(
            angle=angle,
            axis=axis,
            origin=origin,
        )

    def _with_rotation_between_vectors(
        self,
        start: np.ndarray,
        target: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Modify molecule.

        """

        # Set the origin of the rotation to "origin".
        self._with_displacement(-origin)
        rot_mat = rotation_matrix(start, target)

        # Apply the rotation matrix to the atomic positions to yield
        # the new atomic positions.
        self._position_matrix = rot_mat @ self._position_matrix

        # Restore original position.
        self._with_displacement(origin)
        return self

    def with_rotation_between_vectors(
        self,
        start: np.ndarray,
        target: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Return a rotated clone.

        The rotation is equal to a rotation from `start` to `target`.

        Given two direction vectors, `start` and `target`, this method
        applies the rotation required transform `start` to `target`
        onto the clone. The rotation occurs about the `origin`.

        For example, if the `start` and `target` vectors
        are 45 degrees apart, a 45 degree rotation will be applied to
        the clone. The rotation will be along the appropriate
        direction.

        The great thing about this method is that you as long as you
        can associate a geometric feature of the molecule with a
        vector, then the clone can be rotated so that this vector is
        aligned with `target`. The defined vector can be virtually
        anything. This means that any geometric feature of the molecule
        can be easily aligned with any arbitrary direction.

        Parameters:

            start:
                A vector which is to be rotated so that it transforms
                into the `target` vector.

            target:
                The vector onto which `start` is rotated.

            origin:
                The point about which the rotation occurs.

        Returns:

            Molecule: A rotated clone.

        """

        return self.clone()._with_rotation_between_vectors(
            start=start,
            target=target,
            origin=origin,
        )

    def _with_rotation_to_minimize_angle(
        self,
        start: np.ndarray,
        target: np.ndarray,
        axis: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        # If the vector being rotated is not finite, exit. This is
        # probably due to a planar molecule.
        if not all(np.isfinite(x) for x in start):
            return self
        if np.allclose(target, [0, 0, 0], atol=1e-15):
            raise ValueError(
                "target has a magnitude of 0. It is therefore not "
                "possible to calculate an angle."
            )

        self._with_displacement(-origin)

        # 1. Remove any component of the start and target vectors long
        # the axis. This puts them both on the same plane.
        # 2. Calculate the angle between them.
        # 3. Apply the rotation.
        tstart = start - np.dot(start, axis) * axis

        # If `tstart` is 0, it is parallel to the rotation axis, stop.
        if np.allclose(tstart, [0, 0, 0], 1e-8):
            self._with_displacement(origin)
            return self

        tend = target - np.dot(target, axis) * axis
        # If `tend` is 0, it is parallel to the rotation axis, stop.
        if np.allclose(tend, [0, 0, 0], 1e-8):
            self._with_displacement(origin)
            return self

        angle = vector_angle(tstart, tend)

        projection = tstart @ np.cross(axis, tend)
        if projection > 0:
            angle = 2 * np.pi - angle

        rotation_matrix = rotation_matrix_arbitrary_axis(angle, axis)
        self._position_matrix = rotation_matrix @ self._position_matrix
        self._with_displacement(origin)
        return self

    def with_rotation_to_minimize_angle(
        self,
        start: np.ndarray,
        target: np.ndarray,
        axis: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Return a rotated clone.

        The clone is rotated by the rotation required to minimize
        the angle between `start` and `target`.

        Note that this function will not necessarily overlay the
        `start` and `target` vectors. This is because the possible
        rotation is restricted to the `axis`.

        Parameters:

            start:
                The vector which is rotated.

            target:
                The vector which is stationary.

            axis:
                The vector about which the rotation happens. Must have
                unit magnitude.

            origin:
                The origin about which the rotation happens.

        Returns:

            Molecule: A rotated clone.

        Raises:

            :class:`ValueError`
                If `target` has a magnitude of 0. In this case it is
                not possible to calculate an angle between `start` and
                `target`.

        """

        return self.clone()._with_rotation_to_minimize_angle(
            start=start,
            target=target,
            axis=axis,
            origin=origin,
        )

    def clone(self) -> typing.Self:
        """
        Return a clone.

        Returns:
            Molecule: The clone.

        """

        clone = self.__class__.__new__(self.__class__)
        Molecule.__init__(
            self=clone,
            atoms=self._atoms,
            bonds=self._bonds,
            position_matrix=self._position_matrix.T,
        )
        return clone

    def get_atomic_positions(
        self,
        atom_ids: int | Iterable[int] | None = None,
    ) -> Iterator[np.ndarray]:
        """
        Yield the positions of atoms.

        Parameters:

            atom_ids (int | list[int] | None):
                The ids of the atoms whose positions are desired.
                If ``None``, all atoms are used.

        Yields:

            The x, y and z coordinates of an atom.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        positions = self._position_matrix[:, atom_ids].T
        for atomic_position in positions:
            yield atomic_position

    def get_atoms(
        self,
        atom_ids: int | Iterable[int] | None = None,
    ) -> Iterator[Atom]:
        """
        Yield the atoms in the molecule, ordered by id.

        Parameters:

            atom_ids (int | list[int] | None):
                The ids of atoms to yield. If ``None``, all
                atoms are used.

        Yields:

            An atom in the molecule.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)

        for atom_id in atom_ids:
            yield self._atoms[atom_id]

    def get_num_atoms(self) -> int:
        """
        Return the number of atoms in the molecule.

        Returns:

            The number of atoms in the molecule.

        """

        return len(self._atoms)

    def get_bonds(self) -> Iterator[Bond]:
        """
        Yield the bond in the molecule.

        Yields:

            A bond in the molecule.

        """

        for bond in self._bonds:
            yield bond

    def get_num_bonds(self) -> int:
        """
        Return the number of bonds in the molecule.

        Returns:

            The number of bonds in the molecule.

        """

        return len(self._bonds)

    def get_centroid(
        self,
        atom_ids: int | Iterable[int] | None = None,
    ) -> np.ndarray:
        """
        Return the centroid.

        Parameters:

            atom_ids (int | list[int] | None):
                The ids of atoms which are used to calculate the
                centroid. If ``None``, all atoms are used.

        Returns:

            The centroid of atoms specified by `atom_ids`.

        Raises:

            :class:`ValueError`
                If `atom_ids` has a length of ``0``.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        if len(atom_ids) == 0:
            raise ValueError("atom_ids was of length 0.")

        return np.divide(
            self._position_matrix[:, atom_ids].sum(axis=1),
            len(atom_ids),
        )

    def get_direction(
        self,
        atom_ids: int | Iterable[int] | None = None,
    ) -> np.ndarray:
        """
        Return a vector of best fit through the atoms.

        Parameters:

            atom_ids (int | list[int] | None):
                The ids of atoms which should be used to calculate the
                vector. If ``None``, all atoms are used.

        Returns:

            The vector of best fit.

        Raises:

            :class:`ValueError`
                If `atom_ids` has a length of ``0``.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        if len(atom_ids) == 0:
            raise ValueError("atom_ids was of length 0.")

        pos = self._position_matrix[:, atom_ids].T
        return np.around(
            a=np.linalg.svd(pos - pos.mean(axis=0))[-1][0],
            decimals=14,
        )

    def get_maximum_diameter(
        self,
        atom_ids: int | Iterable[int] | None = None,
    ) -> float:
        """
        Return the maximum diameter.

        This method does not account for the van der Waals radius of
        atoms.

        Parameters:

            atom_ids (int | list[int] | None):
                The ids of atoms which are considered when looking for
                the maximum diameter. If ``None``, all atoms are used.

        Returns:

            The maximum diameter in the molecule.

        Raises:

            :class:`ValueError`
                If `atom_ids` has a length of ``0``.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        if len(atom_ids) == 0:
            raise ValueError("atom_ids was of length 0.")

        coords = self._position_matrix[:, atom_ids]
        return float(euclidean(coords.min(axis=1), coords.max(axis=1)))

    def get_plane_normal(
        self,
        atom_ids: int | Iterable[int] | None = None,
    ) -> np.ndarray:
        """
        Return the normal to the plane of best fit.

        Parameters:

            atom_ids (int | list[int] | None):
                The ids of atoms which should be used to calculate the
                plane. If ``None``, all atoms are used.

        Returns:

            Vector orthonormal to the plane of the molecule.

        Raises:

            :class:`ValueError`
                If `atom_ids` has a length of ``0``.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        if len(atom_ids) == 0:
            raise ValueError("atom_ids was of length 0.")

        pos = self._position_matrix[:, atom_ids].T
        centroid = self.get_centroid(atom_ids)
        return np.around(np.linalg.svd(pos - centroid)[-1][2, :], 14)

    def get_position_matrix(self) -> np.ndarray:
        """
        Return a matrix holding the atomic positions.

        Returns:

            The array has the shape ``(n, 3)``. Each row holds the
            x, y and z coordinates of an atom.

        """

        return np.array(self._position_matrix.T)

    def _with_centroid(
        self,
        position: np.ndarray,
        atom_ids: int | Iterable[int] | None,
    ) -> typing.Self:
        centroid = self.get_centroid(atom_ids=atom_ids)
        self._with_displacement(position - centroid)
        return self

    def with_centroid(
        self,
        position: np.ndarray,
        atom_ids: int | Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Return a clone with its centroid at `position`.

        Parameters:

            position:
                This array holds the position on which the centroid of
                the clone is going to be placed.

            atom_ids (int | list[int] | None):
                The ids of atoms which should have their centroid set
                to `position`. If ``None``, all atoms are used.

        Returns:

            Molecule: A clone with its centroid at `position`.

        """

        return self.clone()._with_centroid(position, atom_ids)

    def _with_position_matrix(
        self,
        position_matrix: np.ndarray,
    ) -> typing.Self:
        """
        Modify molecule.

        """

        self._position_matrix = np.array(position_matrix.T)
        return self

    def with_position_matrix(self, position_matrix: np.ndarray) -> typing.Self:
        """
        Return a clone with atomic positions set by `position_matrix`.

        Parameters:

            position_matrix:
                The position matrix of the clone. The shape of the
                matrix is ``(n, 3)``.

        Returns:

            Molecule: The clone.

        """
        return self.clone()._with_position_matrix(position_matrix)

    def to_rdkit_mol(self) -> rdkit.Mol:
        """
        Return an :mod:`rdkit` representation.

        Returns:

            The molecule in :mod:`rdkit` format.

        """

        mol = rdkit.EditableMol(rdkit.Mol())
        for atom in self._atoms:
            rdkit_atom = rdkit.Atom(atom.get_atomic_number())
            rdkit_atom.SetFormalCharge(atom.get_charge())
            mol.AddAtom(rdkit_atom)

        for bond in self._bonds:
            mol.AddBond(
                beginAtomIdx=bond.get_atom1().get_id(),
                endAtomIdx=bond.get_atom2().get_id(),
                order=(
                    rdkit.BondType.DATIVE
                    if bond.get_order() == 9
                    else rdkit.BondType(bond.get_order())
                ),
            )

        mol = mol.GetMol()
        rdkit_conf = rdkit.Conformer(len(self._atoms))
        for atom_id, atom_coord in enumerate(self._position_matrix.T):
            rdkit_conf.SetAtomPosition(atom_id, atom_coord)
            mol.GetAtomWithIdx(atom_id).SetNoImplicit(True)
        mol.AddConformer(rdkit_conf)
        return mol

    def to_atomlite(self) -> atomlite.Molecule:
        """
        Return an :mod:`atomlite` representation.

        Returns:

            The molecule in :mod:`atomlite` JSON format.

        """
        return atomlite.json_from_rdkit(self.to_rdkit_mol())

    def with_structure_from_file(
        self,
        path: pathlib.Path | str,
        extension: str | None = None,
    ) -> typing.Self:
        """
        Return a clone, with its structure taken from a file.

        Multiple file types are supported, namely:

        #. ``.mol``, ``.sdf`` - MDL V2000 and V3000 files
        #. ``.xyz`` - XYZ files
        #. ``.mae`` - Schrodinger Maestro files
        #. ``.coord`` - Turbomole files
        #. ``.pdb`` - PDB files

        Parameters:

            path:
                The path to a molecular structure file holding updated
                coordinates for the :class:`.Molecule`.

            extension:
                If you want to treat the file as though it has a
                particular extension, put it here. Include the dot.

        Returns:

            Molecule: A clone with atomic positions found in `path`.

        """

        if extension is None:
            _, extension = os.path.splitext(path)

        return {
            ".mol": mdl_mol._with_structure_from_mol,
            ".sdf": mdl_mol._with_structure_from_mol,
            ".mae": mae._with_structure_from_mae,
            ".xyz": xyz._with_structure_from_xyz,
            ".coord": turbomole._with_structure_from_turbomole,
            ".pdb": pdb._with_structure_from_pdb,
        }[extension](self.clone(), str(path))

    def with_canonical_atom_ordering(self) -> typing.Self:
        """
        Return a clone, with canonically ordered atoms.

        Returns:
            Molecule: The clone.
        """
        return self.clone()._with_canonical_atom_ordering()

    def _with_canonical_atom_ordering(self) -> typing.Self:
        """
        Modify the molecule.

        """

        atom_map = {
            atom.get_id(): atom.with_id(new_id)
            for new_id, atom in zip(
                rdkit.CanonicalRankAtoms(self.to_rdkit_mol()),
                self._atoms,
            )
        }
        self._atoms = tuple(
            sorted(atom_map.values(), key=lambda atom: atom.get_id())
        )
        self._bonds = tuple(
            sorted(
                (
                    sort_bond_atoms_by_id(bond.with_atoms(atom_map))
                    for bond in self._bonds
                ),
                key=get_bond_atom_ids,
            )
        )
        old_ids = {atom.get_id(): old_id for old_id, atom in atom_map.items()}
        self._position_matrix = np.array(
            np.array(
                [
                    self._position_matrix.T[old_ids[new_id]]
                    for new_id in range(len(self._atoms))
                ]
            ).T
        )
        return self

    def get_canonical_atom_ids(self) -> dict[int, int]:
        """
        Map the id of each atom to its id under canonical ordering.

        Returns:

            Maps the id of each atom in the molecule to the id it would
            have under canonical ordering.

        """

        return {
            old_id: new_id
            for old_id, new_id in enumerate(
                rdkit.CanonicalRankAtoms(self.to_rdkit_mol()),
            )
        }

    def write(
        self,
        path: pathlib.Path | str,
        atom_ids: int | Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Write the structure to a file.

        This function will write the format based on the extension
        of `path`.

        #. ``.mol``, ``.sdf`` - MDL V3000 MOL file
        #. ``.xyz`` - XYZ file
        #. ``.pdb`` - PDB file

        Parameters:

            path:
                The `path` to which the molecule should be written.

            atom_ids (int | list[int] | None):
                The atom ids of atoms to write. If ``None``,
                all atoms are used. If you use this parameter, the
                atom ids in the file may not correspond to the atom
                ids in the molecule.

        Returns:

            Molecule: The molecule.

        """

        _, extension = os.path.splitext(path)
        {
            ".mol": _write_mdl_mol_file,
            ".sdf": _write_mdl_mol_file,
            ".xyz": _write_xyz_file,
            ".pdb": _write_pdb_file,
        }[extension](self, path, atom_ids)
        return self

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} at {id(self)}>"



import typing
from collections.abc import Iterator, Sequence

from stk._internal.ea.crossover.record import CrossoverRecord

T = typing.TypeVar("T")


class MoleculeCrosser(typing.Protocol[T]):
    """
    Performs crossover operations.

    Crossers take multiple molecules and recombine them to make
    new, offspring, molecules.

    Examples:

        *Implementing crossover operations*

        You only need to implement :meth:`.cross`. The source code of any
        of the classes listed in :mod:`.crosser` can serve as good
        examples.

    """

    def cross(self, records: Sequence[T]) -> Iterator[CrossoverRecord[T]]:
        """
        Cross `records`.

        Parameters:
            records (list[T]):
                The molecule records on which a crossover operation is
                performed.

        Yields:
            A record of a crossover operation.

        """
        ...



import mchammer as mch

from stk._internal.construction_state.construction_state import (
    ConstructionState,
)

from .optimizer import Optimizer
from .utilities import get_long_bond_ids, get_mch_bonds, get_subunits


class MCHammer(Optimizer):
    """
    Performs Monte Carlo optimisation of long-bonds in molecules.

    Examples:

        *Structure Optimization*

        Using :class:`.MCHammer` will lead to :class:`.ConstructedMolecule`
        structures without long bonds.

        .. testcode:: structure-optimization

            import stk

            bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
            bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])

            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(bb1, bb2),
                    repeating_unit='AB',
                    num_repeating_units=6,
                    optimizer=stk.MCHammer(),
                ),
            )

        Optimisation with :mod:`stk` simply collects the final position
        matrix. The optimisation's trajectory can be output using the
        :mod:`MCHammer` implementation if required by the user
        [#mchammer]_.

        The open-source optimization code :mod:`MCHammer` specializes in
        the `collapsing` of molecules with long bonds like those
        constructed by :mod:`stk`. This code is entirely nonphysical and
        is, therefore, completely general to any chemistry.

    References:

        .. [#mchammer] https://github.com/andrewtarzia/MCHammer

    """

    def __init__(
        self,
        step_size: float = 0.25,
        target_bond_length: float = 1.2,
        num_steps: int = 500,
        bond_epsilon: float = 50,
        nonbond_epsilon: float = 20,
        nonbond_sigma: float = 1.2,
        nonbond_mu: float = 3,
        beta: float = 2,
        random_seed: int | None = 1000,
    ) -> None:
        """
        Parameters:

            step_size:
                The relative size of the step to take during step.

            target_bond_length:
                Target equilibrium bond length for long bonds to minimize
                to in Angstrom.

            num_steps:
                Number of MC moves to perform.

            bond_epsilon:
                Value of epsilon used in the bond potential in MC moves.
                Determines strength of the bond potential.

            nonbond_epsilon:
                Value of epsilon used in the nonbond potential in MC moves.
                Determines strength of the nonbond potential.
                Larger values lead to a larger building block repulsion.

            nonbond_sigma:
                Value of sigma used in the nonbond potential in MC moves.
                Larger values lead to building block repulsion at larger
                distances.

            nonbond_mu:
                Value of mu used in the nonbond potential in MC moves.
                Determines the steepness of the nonbond potential.

            beta:
                Value of beta used in the in MC moves. Beta takes the
                place of the inverse Boltzmann temperature.

            random_seed:
                Random seed to use for MC algorithm. If
                ``None`` a system-based random seed will be used
                and results will not be reproducible between
                invocations.
        """
        self._optimizer = mch.Optimizer(
            step_size=step_size,
            target_bond_length=target_bond_length,
            num_steps=num_steps,
            bond_epsilon=bond_epsilon,
            nonbond_epsilon=nonbond_epsilon,
            nonbond_sigma=nonbond_sigma,
            nonbond_mu=nonbond_mu,
            beta=beta,
            random_seed=random_seed,
        )

    def optimize(self, state: ConstructionState) -> ConstructionState:
        mch_mol = mch.Molecule(
            atoms=(
                mch.Atom(
                    id=atom.get_id(),
                    element_string=atom.__class__.__name__,
                )
                for atom in state.get_atoms()
            ),
            bonds=tuple(get_mch_bonds(state)),
            position_matrix=state.get_position_matrix(),
        )
        mch_mol, _ = self._optimizer.get_result(
            mol=mch_mol,
            bond_pair_ids=tuple(get_long_bond_ids(state)),
            subunits=get_subunits(state),
        )
        return state.with_position_matrix(
            position_matrix=mch_mol.get_position_matrix(),
        )



from functools import partial

import pytest
import stk
from scipy.spatial.distance import euclidean

from ....case_data import CaseData


@pytest.fixture
def bidentate(position, building_block_2):
    point1, point2 = points = (
        position + [-10, 0, 0],
        position + [0, -10, 0],
    )

    def get_fg0_point(building_block):
        return get_closest_point(
            points=points,
            point=get_fg_position(0, building_block),
        )

    def get_fg1_point(building_block):
        return get_closest_point(
            points=points,
            point=get_fg_position(1, building_block),
        )

    vertex = stk.metal_complex.BiDentateLigandVertex(
        id=0,
        position=position,
    )

    return CaseData(
        vertex=vertex,
        edges=tuple(get_edges(vertex)),
        building_block=building_block_2,
        position=position,
        alignment_tests={
            get_fg0_point: point1,
            get_fg1_point: point2,
        },
        functional_group_edges=({0: 0, 1: 1}),
    )


def get_closest_point(points, point):
    return min(points, key=partial(euclidean, point))


def get_fg_position(id, building_block):
    functional_group = next(building_block.get_functional_groups(id))
    return building_block.get_centroid(
        atom_ids=functional_group.get_placer_ids(),
    )


def get_edges(vertex):
    vertex2 = stk.Vertex(1, vertex.get_position() + [-1, -1, 0])
    yield stk.Edge(
        id=0,
        vertex1=vertex,
        vertex2=vertex2,
        position=vertex.get_position() + [-1, 0, 0],
    )
    yield stk.Edge(
        id=1,
        vertex1=vertex,
        vertex2=vertex2,
        position=vertex.get_position() + [0, -1, 0],
    )


@pytest.fixture(
    params=(
        lambda: stk.BuildingBlock(
            smiles="C=NC/C=N/Br",
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts="[#6]~[#7X2]~[#35]",
                    bonders=(1,),
                    deleters=(),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts="[#6]~[#7X2]~[#6]",
                    bonders=(1,),
                    deleters=(),
                ),
            ],
        ),
    ),
)
def building_block_2(request) -> stk.BuildingBlock:
    return request.param()



import rdkit.Chem.AllChem as rdkit


def test_get_atoms(case_data):
    """
    Test :meth:`.Molecule.get_atoms`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and the correct
        SMILES.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_atoms(case_data.molecule, case_data.smiles)


def _test_get_atoms(molecule, smiles):
    """
    Test :meth:`.Molecule.get_atoms`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    smiles : :class:`str`
        The correct SMILES of `molecule`.

    Returns
    -------
    None : :class:`NoneType`

    """

    # get_rdkit_mol() tests get_atoms() and get_bonds(), if those are
    # broken, the produced rdkit molecule will not have the expected
    # smiles.
    result = rdkit.MolToSmiles(get_rdkit_mol(molecule))
    # Printing is useful for debugging.
    print(result)
    assert result == smiles


def get_rdkit_mol(molecule):
    rdkit_mol = rdkit.EditableMol(rdkit.Mol())
    for atom in molecule.get_atoms():
        rdkit_atom = rdkit.Atom(atom.get_atomic_number())
        rdkit_atom.SetFormalCharge(atom.get_charge())
        rdkit_mol.AddAtom(rdkit_atom)

    for bond in molecule.get_bonds():
        rdkit_mol.AddBond(
            beginAtomIdx=bond.get_atom1().get_id(),
            endAtomIdx=bond.get_atom2().get_id(),
            order=(
                rdkit.BondType.DATIVE
                if bond.get_order() == 9
                else rdkit.BondType(bond.get_order())
            ),
        )
    return rdkit_mol.GetMol()



def test_get_bond(case_data):
    """
    Test :meth:`.BondInfo.get_bond`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the bond info to test and the bond it
        should be holding.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.bond_info.get_bond() is case_data.bond



from stk._internal.topology_graphs.small.internal_reaction import (
    InternalReaction,
)
from stk._internal.topology_graphs.small.ncore import NCore
from stk._internal.topology_graphs.small.vertices import (
    CoreVertex,
    SingleVertex,
    SubstituentVertex,
)

__all__ = [
    "NCore",
    "CoreVertex",
    "SubstituentVertex",
    "SingleVertex",
    "InternalReaction",
]



import pytest

import stk

from ...case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.small.NCore(
                    core_building_block=(
                        stk.BuildingBlock(
                            smiles="BrC(Br)Br",
                            functional_groups=stk.BromoFactory(),
                        )
                    ),
                    arm_building_blocks=stk.BuildingBlock(
                        smiles="BrC",
                        functional_groups=stk.BromoFactory(),
                    ),
                    repeating_unit="A",
                ),
            ),
            smiles="[H]C([H])([H])C([H])(C([H])([H])[H])C([H])([H])[H]",
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.small.NCore(
                    core_building_block=(
                        stk.BuildingBlock(
                            smiles="C(Br)1C(Br)C(Br)C(Br)C(Br)C(Br)C1Br",
                            functional_groups=stk.BromoFactory(),
                        )
                    ),
                    arm_building_blocks=stk.BuildingBlock(
                        smiles="BrC",
                        functional_groups=stk.BromoFactory(),
                    ),
                    repeating_unit="A",
                ),
            ),
            smiles=(
                "[H]C([H])([H])C1([H])C([H])(C([H])([H])[H])C([H])(C([H"
                "])([H])[H])C([H])(C([H])([H])[H])C([H])(C([H])([H])[H]"
                ")C([H])(C([H])([H])[H])C1([H])C([H])([H])[H]"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.small.NCore(
                    core_building_block=(
                        stk.BuildingBlock(
                            smiles="C(Br)1C(Br)C(Br)C(Br)C(Br)C(Br)C1Br",
                            functional_groups=stk.BromoFactory(),
                        )
                    ),
                    arm_building_blocks=[
                        stk.BuildingBlock(
                            smiles="BrC",
                            functional_groups=stk.BromoFactory(),
                        ),
                        stk.BuildingBlock(
                            smiles="BrCN",
                            functional_groups=stk.BromoFactory(),
                        ),
                    ],
                    repeating_unit="ABABABA",
                ),
            ),
            smiles=(
                "[H]N([H])C([H])([H])C1([H])C([H])(C([H])([H])[H])C([H]"
                ")(C([H])([H])[H])C([H])(C([H])([H])N([H])[H])C([H])(C("
                "[H])([H])[H])C([H])(C([H])([H])N([H])[H])C1([H])C([H])"
                "([H])[H]"
            ),
            name=name,
        ),
    ),
)
def small_ncore(request: pytest.FixtureRequest) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



"""
Deletions Summary
=================

"""

import numpy as np

from stk._internal.atom_info import AtomInfo
from stk._internal.bond_info import BondInfo


class _DeletionsSummary:
    """
    A summary of deletion results.

    """

    __slots__ = [
        "_atoms",
        "_atom_infos",
        "_bonds",
        "_bond_infos",
        "_position_matrix",
        "_deleted_atom_ids",
        "_deleted_bond_ids",
        "_valid_atoms",
        "_valid_atom_infos",
        "_valid_bonds",
        "_valid_bond_infos",
        "_valid_positions",
    ]

    def __init__(
        self,
        atoms,
        atom_infos,
        bonds,
        bond_infos,
        position_matrix,
        deleted_atom_ids,
        deleted_bond_ids,
    ):
        """
        Initialize a :class:`._DeletionsSummary` instance.

        Parameters
        ----------
        atoms : :class:`iterable` of :class:`.Atom`
            Atoms, some of which are to be deleted.

        atom_infos : :class:`iterable` of :class:`.AtomInfo`
            Info on every atom in `atoms`.

        bonds : :class:`iterable` of :class:`.Bond`
            The bonds of the molecule being constructed.

        bond_infos : :class:`iterable` of :class:`.BondInfo`
            Info on every bond in `bonds`.

        position_matrix : :class:`numpy.ndarray`
            The position matrix for all the `atoms`.

        deleted_atom_ids : :class:`iterable` of :class:`int`
            The ids of `atoms`, which should be deleted.

        deleted_bond_ids : :class:`iterable` of :class:`._BondId`
            Ids of bonds that should be deleted.

        """

        self._atoms = tuple(atoms)
        self._atom_infos = tuple(atom_infos)
        self._bonds = tuple(bonds)
        self._bond_infos = tuple(bond_infos)
        self._position_matrix = np.array(position_matrix)
        self._deleted_atom_ids = set(deleted_atom_ids)
        self._deleted_bond_ids = set(deleted_bond_ids)

        self._valid_atoms = []
        self._valid_atom_infos = []
        self._valid_bonds = []
        self._valid_bond_infos = []
        self._valid_positions = []
        self._with_valid_data()

    def _with_valid_data(self):
        """
        Add the data of atoms which aren't deleted.

        Returns
        -------
        None : :class:`NoneType`

        """

        atoms = self._atoms
        atom_infos = self._atom_infos
        position_matrix = self._position_matrix
        deleted_atom_ids = self._deleted_atom_ids
        deleted_bond_ids = self._deleted_bond_ids

        def valid_atom(atom):
            return atom.get_id() not in deleted_atom_ids

        valid_atoms = self._valid_atoms
        valid_atom_infos = self._valid_atom_infos
        valid_positions = self._valid_positions
        atom_map = {}

        def with_atom(atom):
            atom_id = atom.get_id()
            valid_atoms.append(atom.with_id(len(valid_atoms)))
            atom_map[atom_id] = valid_atoms[-1]
            info = atom_infos[atom_id]
            valid_atom_infos.append(
                AtomInfo(
                    atom=valid_atoms[-1],
                    building_block_atom=info.get_building_block_atom(),
                    building_block=info.get_building_block(),
                    building_block_id=info.get_building_block_id(),
                )
            )
            valid_positions.append(position_matrix[atom_id])

        for atom in filter(valid_atom, atoms):
            with_atom(atom)

        def valid_bond(bond_data):
            index, bond = bond_data
            atom1_id = bond.get_atom1().get_id()
            atom2_id = bond.get_atom2().get_id()
            return (
                atom1_id not in deleted_atom_ids
                and atom2_id not in deleted_atom_ids
                and (atom1_id, atom2_id) not in deleted_bond_ids
            )

        bonds = self._bonds
        bond_infos = self._bond_infos
        valid_bonds = self._valid_bonds
        valid_bond_infos = self._valid_bond_infos

        def with_bond(index, bond):
            valid_bonds.append(bond.with_atoms(atom_map))
            info = bond_infos[index]
            valid_bond_infos.append(
                BondInfo(
                    bond=valid_bonds[-1],
                    building_block=info.get_building_block(),
                    building_block_id=info.get_building_block_id(),
                )
            )

        for index, bond in filter(valid_bond, enumerate(bonds)):
            with_bond(index, bond)

    def get_atoms(self):
        """
        Yield the atoms in the summary.

        Yields
        ------
        :class:`.Atom`
            An atom.

        """

        yield from self._valid_atoms

    def get_atom_infos(self):
        """
        Yield infos about atoms in the summary.

        Yields
        ------
        :class:`.AtomInfo`
            Info about an atom.

        """

        yield from self._valid_atom_infos

    def get_bonds(self):
        """
        Yield the bonds in the summary.

        Yields
        ------
        :class:`.Bond`
            A bond.

        """

        yield from self._valid_bonds

    def get_bond_infos(self):
        """
        Yield infos about the bonds in the summary.

        Yields
        ------
        :class:`.BondInfo`
            Info about a bond.

        """

        yield from self._valid_bond_infos

    def get_positions(self):
        """
        Yield the positions of atoms held by the summary.

        Yields
        ------
        :class:`numpy.ndarray`
            The position of an atom held by the summary.

        """

        yield from self._valid_positions



from .eight_plus_sixteen import *  # noqa
from .five_plus_ten import *  # noqa
from .four_plus_eight import *  # noqa
from .six_plus_twelve import *  # noqa
from .ten_plus_twenty import *  # noqa
from .three_plus_six import *  # noqa
from .two_plus_four import *  # noqa



from .eight_plus_twelve import *  # noqa
from .four_plus_six import *  # noqa
from .four_plus_six_2 import *  # noqa
from .six_plus_nine import *  # noqa
from .twenty_plus_thirty import *  # noqa
from .two_plus_three import *  # noqa



"""
Dative Reaction Utilities
=========================

"""

from itertools import chain


def is_metal(atom):
    """
    Check if `atom` is a metal atom.

    Parameters
    ----------
    atom : :class:`.Atom`
        An atom.

    Returns
    -------
    :class:`bool`
        ``True`` if `atom` is a metal atom.

    """

    metal_atomic_numbers = chain(
        range(21, 31),
        range(39, 49),
        range(72, 81),
    )
    return atom.get_atomic_number() in metal_atomic_numbers



import itertools as it
from functools import partial

import numpy as np
import pytest
import stk

from ..utilities import is_clone
from .check_atom_infos import check_atom_infos
from .check_atoms import check_atoms
from .check_bond_infos import check_bond_infos
from .check_bonds import check_bonds
from .check_building_block_counts import check_building_block_counts
from .check_building_blocks import check_building_blocks
from .check_edge_functional_groups import check_edge_functional_groups
from .check_edges import check_edges
from .check_position_matrix import check_position_matrix
from .check_vertex_edges import check_vertex_edges
from .check_vertices import check_vertices


@pytest.mark.skip
def test_with_placement_results(construction_state):
    building_blocks = tuple(
        map(
            construction_state.get_building_block,
            range(construction_state.get_num_vertices()),
        )
    )
    placement_results = tuple(
        map(
            partial(get_placement_result, construction_state),
            building_blocks,
        )
    )

    # Clone for testing immutability.
    clone = construction_state.clone()
    new_state = construction_state.with_placement_results(
        building_blocks=building_blocks,
        results=placement_results,
    )
    old_state = construction_state
    check_atoms(old_state, new_state, building_blocks)
    check_atom_infos(old_state, new_state, building_blocks)
    check_bonds(old_state, new_state, building_blocks)
    check_bond_infos(old_state, new_state, building_blocks)
    check_building_blocks(old_state, new_state)
    check_building_block_counts(old_state, new_state, building_blocks)
    check_edges(old_state, new_state)
    check_vertex_edges(old_state, new_state)
    check_position_matrix(old_state, new_state, placement_results)
    check_vertices(old_state, new_state)
    check_edge_functional_groups(
        old_state=old_state,
        new_state=new_state,
        building_blocks=building_blocks,
        placement_results=placement_results,
    )
    is_clone(construction_state, clone)


def get_placement_result(construction_state, building_block):
    edges = it.cycle(range(construction_state.get_num_edges()))
    return stk.PlacementResult(
        position_matrix=np.array(
            [[i, i, i] for i in range(building_block.get_num_atoms())],
            dtype=np.float64,
        ),
        functional_group_edges={
            fg_id: next(edges)
            for fg_id in range(building_block.get_num_functional_groups())
        },
    )



import pytest
from pytest_lazyfixture import lazy_fixture

# Fixtures must be visible for lazy_fixture() calls.
from .nrotaxane import *  # noqa


@pytest.fixture(
    params=(lazy_fixture("rotaxane_nrotaxane"),),
)
def rotaxane(request):
    return request.param



from functools import partial

import pytest
import stk
from scipy.spatial.distance import euclidean

from ....case_data import CaseData


@pytest.fixture
def linear(position, linear_aligner_edge, building_block_2):
    point1, point2 = points = (
        position + [-10, 0, 0],
        position + [10, 0, 0],
    )

    def get_fg0_point(building_block):
        return get_closest_point(
            points=points,
            point=get_fg_position(0, building_block),
        )

    def get_fg1_point(building_block):
        return get_closest_point(
            points=points,
            point=get_fg_position(1, building_block),
        )

    vertex = stk.cof.LinearVertex(
        id=0,
        position=position,
        aligner_edge=linear_aligner_edge,
        cell=[0, 0, 0],
    )

    return CaseData(
        vertex=vertex,
        edges=tuple(get_linear_edges(vertex)),
        building_block=building_block_2,
        position=position,
        alignment_tests={
            get_fg0_point: (point1 if linear_aligner_edge == 0 else point2),
            get_fg1_point: (point2 if linear_aligner_edge == 0 else point1),
        },
        functional_group_edges=(
            {0: 0, 1: 1} if linear_aligner_edge == 0 else {0: 1, 1: 0}
        ),
    )


def get_closest_point(points, point):
    return min(points, key=partial(euclidean, point))


def get_fg_position(id, building_block):
    functional_group = next(building_block.get_functional_groups(id))
    return building_block.get_centroid(
        atom_ids=functional_group.get_placer_ids(),
    )


def get_linear_edges(vertex):
    vertex2 = stk.Vertex(1, vertex.get_position() + [-10, 0, 0])
    vertex3 = stk.Vertex(2, vertex.get_position() + [10, 0, 0])
    yield stk.cof.Edge(0, 0, vertex, vertex2)
    yield stk.cof.Edge(1, 1, vertex, vertex3)


@pytest.fixture(params=(0, 1))
def linear_aligner_edge(request):
    return request.param


@pytest.fixture(
    params=(
        lambda: stk.BuildingBlock(
            smiles="BrCCNCBr",
            functional_groups=[stk.BromoFactory()],
        ),
    ),
)
def building_block_2(request) -> stk.BuildingBlock:
    return request.param()



import numpy as np
import stk


def test_put_caching(mongo_client):
    database_name = "_test_put_caching"
    mongo_client.drop_database(database_name)

    database = stk.ConstructedMoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
    )
    molecule = stk.BuildingBlock("BrCCCBr", [stk.BromoFactory()])
    polymer = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            building_blocks=(molecule,),
            repeating_unit="A",
            num_repeating_units=3,
        ),
    )
    database.put(polymer)
    database.put(polymer)

    cache_info = database._put.cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 1

    database.put(
        molecule=polymer.with_position_matrix(
            position_matrix=np.zeros((polymer.get_num_atoms(), 3)),
        ),
    )
    cache_info = database._put.cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 2


def test_get_caching(mongo_client):
    database_name = "_test_get_caching"
    mongo_client.drop_database(database_name)

    database = stk.ConstructedMoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
    )
    molecule = stk.BuildingBlock("BrCCCBr", [stk.BromoFactory()])
    polymer = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            building_blocks=(molecule,),
            repeating_unit="A",
            num_repeating_units=3,
        ),
    )
    database.put(polymer)
    database.get(
        {
            stk.InchiKey().get_key_name(): stk.InchiKey().get_key(polymer),
        }
    )
    database.get(
        {
            stk.InchiKey().get_key_name(): stk.InchiKey().get_key(polymer),
        }
    )

    cache_info = database._get.cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 1



"""
Boronic Acid
============

"""

from __future__ import annotations

import typing

from stk._internal.atom import Atom
from stk._internal.elements import B, H, O

from .generic_functional_group import GenericFunctionalGroup


class BoronicAcid(GenericFunctionalGroup):
    """
    Represents a boronic acid functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[atom][boron]([oxygen1][hydrogen1])[oxygen2][hydrogen2]``.

    """

    def __init__(
        self,
        boron: B,
        oxygen1: O,  # noqa: Not an ambiguous name.
        hydrogen1: H,
        oxygen2: O,  # noqa: Not an ambiguous name.
        hydrogen2: H,
        atom: Atom,
        bonders: tuple[Atom, ...],
        deleters: tuple[Atom, ...],
        placers: typing.Optional[tuple[Atom, ...]] = None,
    ) -> None:
        """
        Initialize a :class:`.BoronicAcid` instance.

        Parameters:

            boron:
                The ``[boron]`` atom.

            oxygen1:
                The ``[oxygen1]`` atom.

            hydrogen1:
                The ``[hydrogen]`` atom.

            oxygen2:
                The ``[oyxgen2]`` atom.

            hydrogen2:
                The ``[hydrogen2]`` atom.

            atom:
                The ``[atom]`` atom.

            bonders:
                The bonder atoms.

            deleters:
                The deleter atoms.

            placers:
                The placer atoms. If ``None`` the `bonders` will be
                used.

        """

        self._boron = boron
        self._oxygen1 = oxygen1
        self._hydrogen1 = hydrogen1
        self._oxygen2 = oxygen2
        self._hydrogen2 = hydrogen2
        self._atom = atom
        atoms = (boron, oxygen1, hydrogen1, oxygen2, hydrogen2, atom)
        super().__init__(
            atoms=atoms,
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_boron(self) -> B:
        """
        Get the ``[boron]`` atom.

        Returns:

            The ``[boron]`` atom.

        """

        return self._boron

    def get_oxygen1(self) -> O:  # noqa: Not an ambiguous name.
        """
        Get the ``[oxygen1]`` atom.

        Returns:

            The ``[oxygen1]`` atom.

        """

        return self._oxygen1

    def get_hydrogen1(self) -> H:
        """
        Get the ``[hydrogen1]`` atom.

        Returns:

            The ``[hydrogen1]`` atom.

        """

        return self._hydrogen1

    def get_oxygen2(self) -> O:  # noqa: Not an ambiguous name.
        """
        Get the ``[oxygen2]`` atom.

        Returns:

            The ``[oxygen2]`` atom.

        """

        return self._oxygen2

    def get_hydrogen2(self) -> H:
        """
        Get the ``[hydrogen2]`` atom.

        Returns:

            The ``[hydrogen2]`` atom.

        """

        return self._hydrogen2

    def get_atom(self) -> Atom:
        """
        Get the ``[atom]`` atom.

        Returns:

            The ``[atom]`` atom.

        """

        return self._atom

    def clone(self) -> BoronicAcid:
        clone = super()._clone()
        clone._boron = self._boron
        clone._oxygen1 = self._oxygen1
        clone._hydrogen1 = self._hydrogen1
        clone._oxygen2 = self._oxygen2
        clone._hydrogen2 = self._hydrogen2
        clone._atom = self._atom
        return clone

    def _with_ids(
        self,
        id_map: dict[int, int],
    ) -> BoronicAcid:
        super()._with_ids(id_map)

        if (boron_id := self._boron.get_id()) in id_map:
            self._boron = self._boron.with_id(id_map[boron_id])

        if (oxygen1_id := self._oxygen1.get_id()) in id_map:
            self._oxygen1 = self._oxygen1.with_id(id_map[oxygen1_id])

        if (hydrogen1_id := self._hydrogen1.get_id()) in id_map:
            self._hydrogen1 = self._hydrogen1.with_id(
                id=id_map[hydrogen1_id],
            )

        if (oxygen2_id := self._oxygen2.get_id()) in id_map:
            self._oxygen2 = self._oxygen2.with_id(id_map[oxygen2_id])

        if (hydrogen2_id := self._hydrogen2.get_id()) in id_map:
            self._hydrogen2 = self._hydrogen2.with_id(
                id=id_map[hydrogen2_id],
            )

        if (atom_id := self._atom.get_id()) in id_map:
            self._atom = self._atom.with_id(id_map[atom_id])

        return self

    def with_ids(
        self,
        id_map: dict[int, int],
    ) -> BoronicAcid:
        return self.clone()._with_ids(id_map)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"{self._boron}, {self._oxygen1}, {self._hydrogen1}, "
            f"{self._oxygen2}, {self._hydrogen2}, {self._atom}, "
            f"bonders={self._bonders}, deleters={self._deleters})"
        )



"""
Utilities
=========


This module defines general-purpose objects, functions and classes.

Functions, classes, etc. defined here should not depend on any other
part of ``stk``. They must be completely self-sufficient.

"""

import os
import re
import tarfile
import time
import typing
from collections import abc, deque
from contextlib import contextmanager

import numpy as np
import rdkit.Chem.AllChem as rdkit
from rdkit.Geometry import Point3D
from scipy.spatial.transform import Rotation

# Holds the elements Van der Waals radii in Angstroms.
atom_vdw_radii = {
    "Al": 2,
    "Sb": 2,
    "Ar": 1.88,
    "As": 1.85,
    "Ba": 2,
    "Be": 2,
    "Bi": 2,
    "B": 2,
    "Br": 1.85,
    "Cd": 1.58,
    "Cs": 2,
    "Ca": 2,
    "C": 1.7,
    "Ce": 2,
    "Cl": 1.75,
    "Cr": 2,
    "Co": 2,
    "Cu": 1.4,
    "Dy": 2,
    "Er": 2,
    "Eu": 2,
    "F": 1.47,
    "Gd": 2,
    "Ga": 1.87,
    "Ge": 2,
    "Au": 1.66,
    "Hf": 2,
    "He": 1.4,
    "Ho": 2,
    "H": 1.09,
    "In": 1.93,
    "I": 1.98,
    "Ir": 2,
    "Fe": 2,
    "Kr": 2.02,
    "La": 2,
    "Pb": 2.02,
    "Li": 1.82,
    "Lu": 2,
    "Mg": 1.73,
    "Mn": 2,
    "Hg": 1.55,
    "Mo": 2,
    "Nd": 2,
    "Ne": 1.54,
    "Ni": 1.63,
    "Nb": 2,
    "N": 1.55,
    "Os": 2,
    "O": 1.52,
    "Pd": 1.63,
    "P": 1.8,
    "Pt": 1.72,
    "K": 2.75,
    "Pr": 2,
    "Pa": 2,
    "Re": 2,
    "Rh": 2,
    "Rb": 2,
    "Ru": 2,
    "Sm": 2,
    "Sc": 2,
    "Se": 1.9,
    "Si": 2.1,
    "Ag": 1.72,
    "Na": 2.27,
    "Sr": 2,
    "S": 1.8,
    "Ta": 2,
    "Te": 2.06,
    "Tb": 2,
    "Tl": 1.96,
    "Th": 2,
    "Tm": 2,
    "Sn": 2.17,
    "Ti": 2,
    "W": 2,
    "U": 1.86,
    "V": 2,
    "Xe": 2.16,
    "Yb": 2,
    "Y": 2,
    "Zn": 1.29,
    "Zr": 2,
    "X": 1.0,
    "D": 1.0,
}

# This dictionary gives easy access to the rdkit bond types.
bond_dict = {
    "1": rdkit.rdchem.BondType.SINGLE,
    "am": rdkit.rdchem.BondType.SINGLE,
    "2": rdkit.rdchem.BondType.DOUBLE,
    "3": rdkit.rdchem.BondType.TRIPLE,
    "ar": rdkit.rdchem.BondType.AROMATIC,
}

# A dictionary which matches atomic number to elemental symbols.
periodic_table = {
    1: "H",
    2: "He",
    3: "Li",
    4: "Be",
    5: "B",
    6: "C",
    7: "N",
    8: "O",
    9: "F",
    10: "Ne",
    11: "Na",
    12: "Mg",
    13: "Al",
    14: "Si",
    15: "P",
    16: "S",
    17: "Cl",
    18: "Ar",
    19: "K",
    20: "Ca",
    21: "Sc",
    22: "Ti",
    23: "V",
    24: "Cr",
    25: "Mn",
    26: "Fe",
    27: "Co",
    28: "Ni",
    29: "Cu",
    30: "Zn",
    31: "Ga",
    32: "Ge",
    33: "As",
    34: "Se",
    35: "Br",
    36: "Kr",
    37: "Rb",
    38: "Sr",
    39: "Y",
    40: "Zr",
    41: "Nb",
    42: "Mo",
    43: "Tc",
    44: "Ru",
    45: "Rh",
    46: "Pd",
    47: "Ag",
    48: "Cd",
    49: "In",
    50: "Sn",
    51: "Sb",
    52: "Te",
    53: "I",
    54: "Xe",
    55: "Cs",
    56: "Ba",
    57: "La",
    58: "Ce",
    59: "Pr",
    60: "Nd",
    61: "Pm",
    62: "Sm",
    63: "Eu",
    64: "Gd",
    65: "Tb",
    66: "Dy",
    67: "Ho",
    68: "Er",
    69: "Tm",
    70: "Yb",
    71: "Lu",
    72: "Hf",
    73: "Ta",
    74: "W",
    75: "Re",
    76: "Os",
    77: "Ir",
    78: "Pt",
    79: "Au",
    80: "Hg",
    81: "Tl",
    82: "Pb",
    83: "Bi",
    84: "Po",
    85: "At",
    86: "Rn",
    87: "Fr",
    88: "Ra",
    89: "Ac",
    90: "Th",
    91: "Pa",
    92: "U",
    93: "Np",
    94: "Pu",
    95: "Am",
    96: "Cm",
    97: "Bk",
    98: "Cf",
    99: "Es",
    100: "Fm",
    101: "Md",
    102: "No",
    103: "Lr",
    104: "Rf",
    105: "Db",
    106: "Sg",
    107: "Bh",
    108: "Hs",
    109: "Mt",
    110: "Ds",
    111: "Rg",
    112: "Cn",
    113: "Uut",
    114: "Fl",
    115: "Uup",
    116: "Lv",
    117: "Uus",
    118: "Uuo",
}


T = typing.TypeVar("T")
OneOrMany = typing.Union[T, abc.Iterable[T]]


class Cell:
    """
    Represents an individual cell in a supercell.

    Parameters
    ----------
    id : :class:`list` of :class:`int`
        A 3 member :class:`list` holding the x, y and z index
        of the cell within the supercell.

    fgs : :class:`dict`
        Maps the fgs in the original unit cell to the
        equivalent fgs in the cell.

    Attributes
    ----------
    id : :class:`numpy.ndarray` of :class:`int`
        A 3 member array holding the x, y and z index
        of the cell within the supercell.

    fgs : :class:`dict`
        Maps the fgs in the original unit cell to the
        equivalent fgs in the cell.

    """

    def __init__(self, id_, fgs):
        self.id = np.array(id_)
        self.fgs = fgs


class ChargedMolError(Exception):
    def __init__(self, mol_file, msg):
        self.mol_file = mol_file
        self.msg = msg


class MolFileError(Exception):
    def __init__(self, mol_file, msg):
        self.mol_file = mol_file
        self.msg = msg


class PopulationSizeError(Exception):
    def __init__(self, msg):
        self.msg = msg


class LazyAttr:
    """
    A descriptor for creating lazy attributes.

    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        val = self.func(obj)
        setattr(obj, self.func.__name__, val)
        return val


def archive_output():
    """
    Places the ``output`` folder into ``stk_ea_runs``.

    This function assumes that the ``output`` folder is in the current
    working directory.

    Returns
    -------
    None : :class:`NoneType`

    """

    if "output" not in os.listdir():
        return

    # Make the ``stk_ea_runs`` folder if it does not exist already.
    if "stk_ea_runs" not in os.listdir():
        os.mkdir("stk_ea_runs")

    # Find out with what number the ``output`` folder should be
    # labelled within ``stk_ea_runs``.
    num = len(os.listdir("stk_ea_runs"))
    new_dir = os.path.join("stk_ea_runs", str(num))
    os.rename("output", new_dir)


def dedupe(iterable, key=None, seen=None):
    """
    Yields items from `iterable` barring duplicates.

    If `seen` is provided it contains elements which are not to be
    yielded at all.

    Parameters
    ----------
    iterable : :class:`iterable`
        An iterable of elements which are to be yielded, only once.

    key : :class:`callable`
        A function which gets applied to every member of `iterable`.
        The return of this function is checked for duplication rather
        than the member itself.

    seen : :class:`set`, optional
        Holds items which are not to be yielded.

    Yields
    ------
    :class:`object`
        Element in `iterable` which is not in `seen` and has not been
        yielded before.

    """

    if seen is None:
        seen = set()
    for x in iterable:
        val = key(x) if key is not None else x
        if val not in seen:
            seen.add(val)
            yield x


def flatten(iterable, excluded_types=None):
    """
    Transforms an nested iterable into a flat one.

    Examples
    --------
    *Flattening a Deeply Nested Structure*

    .. testcode:: flattening-a-deeply-nested-structure

        import stk

        nested = [[1, 2, 3], [[4], [5], [[6]], 7]]
        print(list(stk.flatten(nested)))

    gives

    .. testoutput:: flattening-a-deeply-nested-structure

        [1, 2, 3, 4, 5, 6, 7]

    *Avoiding Flattening of Some Types*

    If a type is found in `excluded_types` it will not be yielded from.
    For example, ``str`` is in `excluded_types` by default, so

    .. testcode:: avoiding-flattening-of-some-types

        import stk

        nested = ['abcd', ['efgh']]
        print(list(stk.flatten(nested)))

    gives

    .. testoutput:: avoiding-flattening-of-some-types

        ['abcd', 'efgh']

    instead of

    .. code-block::

        ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']

    Parameters
    ----------
    iterable : :class:`iterable`
        The iterable which is to be flattened

    excluded_types : :class:`set`, optional
        Holds container types which are not be flattened.

    Yields
    ------
    :class:`object`
        A nested element of `iterable`.

    """

    if excluded_types is None:
        excluded_types = {str}

    for x in iterable:
        if hasattr(x, "__iter__") and type(x) not in excluded_types:
            yield from flatten(x, excluded_types)
        else:
            yield x


def kabsch(coords1, coords2):
    """
    Return a rotation matrix to minimize dstance between 2 coord sets.

    This is essentially an implementation of the Kabsch algorithm.
    Given two sets of coordinates, `coords1` and `coords2`, this
    function returns a rotation matrix. When the rotation matrix is
    applied to `coords1` the resulting coordinates have their rms
    distance to `coords2` minimized.

    Parameters
    ----------
    coords1 : :class:`numpy.ndarray`
        This array represents a matrix hodling coordinates which need
        to be rotated to minimize their rms distance to coordinates in
        `coords2`. The matrix is n x 3. Each row of the matrix holds
        the x, y and z coordinates of one point, respectively. Here
        ``n`` is the number of points.

    coords2 : :class:`numpy.ndarray`
        This array represents a matrix which holds the coordinates of
        points the distance to which should be minimized. The matrix is
        ``n x 3``. Each row of the matrix holds the x, y and z
        coordinates of one point, respectively. Here ``n`` is the
        number of points.

    Returns
    -------
    :class:`numpy.ndarray`
        A rotation matrix. This will be a ``3 x 3`` matrix.

    References
    ----------
    http://nghiaho.com/?page_id=671
    https://en.wikipedia.org/wiki/Kabsch_algorithm

    """

    h = np.dot(coords1, coords2.T)
    u, s, v = np.linalg.svd(h)

    if int(np.linalg.det(v)) < 0:
        v[:, 2] = -v[:, 2]

    return np.dot(v, u)


def matrix_centroid(matrix):
    """
    Returns the centroid of the coordinates held in `matrix`.

    Parameters
    ----------
    matrix : :class:`np.ndarray`
        A ``n x 3`` matrix. Each row holds the x, y and z
        coordinate of some point, respectively.

    Returns
    -------
    :class:`numpy.ndarray`
        A numpy array which holds the x, y and z
        coordinates of the centroid of the coordinates in `matrix`.

    """

    return np.sum(matrix, axis=0) / len(matrix)


def mol_from_mae_file(mae_path):
    """
    Creates a ``rdkit`` molecule from a ``.mae`` file.

    Parameters
    ----------
    mol2_file : :class:`str`
        The full path of the ``.mae`` file from which an rdkit molecule
        should be instantiated.

    Returns
    -------
    :class:`rdkit.Mol`
        An ``rdkit`` instance of the molecule held in `mae_file`.

    """

    mol = rdkit.EditableMol(rdkit.Mol())
    conf = rdkit.Conformer()

    with open(mae_path, "r") as mae:
        content = re.split(r"[{}]", mae.read())

    prev_block = deque([""], maxlen=1)
    for block in content:
        if "m_atom[" in prev_block[0]:
            atom_block = block
        if "m_bond[" in prev_block[0]:
            bond_block = block
        prev_block.append(block)

    labels, data_block, *_ = atom_block.split(":::")
    labels = [
        label
        for label in labels.split("\n")
        if not label.isspace() and label != ""
    ]

    data_block = [
        a.split()
        for a in data_block.split("\n")
        if not a.isspace() and a != ""
    ]

    for line in data_block:
        line = [word for word in line if word != '"']
        if len(labels) != len(line):
            raise RuntimeError(
                (
                    "Number of labels does"
                    " not match number of columns"
                    " in .mae file."
                )
            )

        for label, data in zip(labels, line):
            if "x_coord" in label:
                x = float(data)
            if "y_coord" in label:
                y = float(data)
            if "z_coord" in label:
                z = float(data)
            if "atomic_number" in label:
                atom_num = int(data)

        atom_sym = periodic_table[atom_num]
        atom_coord = Point3D(x, y, z)
        atom_id = mol.AddAtom(rdkit.Atom(atom_sym))
        conf.SetAtomPosition(atom_id, atom_coord)

    labels, data_block, *_ = bond_block.split(":::")
    labels = [
        label
        for label in labels.split("\n")
        if not label.isspace() and label != ""
    ]
    data_block = [
        a.split()
        for a in data_block.split("\n")
        if not a.isspace() and a != ""
    ]

    for line in data_block:
        if len(labels) != len(line):
            raise RuntimeError(
                (
                    "Number of labels does"
                    " not match number of "
                    "columns in .mae file."
                )
            )

        for label, data in zip(labels, line):
            if "from" in label:
                atom1 = int(data) - 1
            if "to" in label:
                atom2 = int(data) - 1
            if "order" in label:
                bond_order = str(int(data))
        mol.AddBond(atom1, atom2, bond_dict[bond_order])

    mol = mol.GetMol()
    mol.AddConformer(conf)
    return mol


def normalize_vector(vector):
    """
    Normalizes the given vector.

    A new vector is returned, the original vector is not modified.

    Parameters
    ----------
    vector : :class:`np.ndarray`
        The vector to be normalized.

    Returns
    -------
    :class:`np.ndarray`
        The normalized vector.

    """

    return np.divide(vector, np.linalg.norm(vector))


def remake(mol):
    """
    Remakes a molecule from scratch.

    Parameters
    ----------
    mol : :class:`rdkit.Mol`
        The molecule to be remade.

    Returns
    -------
    :class:`rdkit.Mol`
        The remade molecule.

    """

    emol = rdkit.EditableMol(rdkit.Mol())
    for atom in mol.GetAtoms():
        new = rdkit.Atom(atom.GetAtomicNum())
        new.SetFormalCharge(atom.GetFormalCharge())
        emol.AddAtom(new)

    for bond in mol.GetBonds():
        emol.AddBond(
            beginAtomIdx=bond.GetBeginAtomIdx(),
            endAtomIdx=bond.GetEndAtomIdx(),
            order=bond.GetBondType(),
        )

    m = emol.GetMol()
    m.AddConformer(mol.GetConformer())
    return m


def get_projection(start, target):
    """
    Get the projection of `start` onto `target`.

    """

    return (
        target
        * np.dot(
            start,
            target,
        )
        / np.dot(target, target)
    )


def orthogonal_vector(vector):
    ortho = np.array([0.0, 0.0, 0.0])
    for m, val in enumerate(vector):
        if not np.allclose(val, 0, atol=1e-8):
            n = (m + 1) % 3
            break
    ortho[n] = vector[m]
    ortho[m] = -vector[n]
    return ortho


def rotation_matrix(vector1, vector2):
    """
    Returns a rotation matrix which transforms `vector1` to `vector2`.

    Multiplying `vector1` by the rotation matrix returned by this
    function yields `vector2`.

    Parameters
    ----------
    vector1 : :class:`numpy.ndarray`
        The vector which needs to be transformed to `vector2`.

    vector2 : :class:`numpy.ndarray`
        The vector onto which `vector1` needs to be transformed.

    Returns
    -------
    :class:`numpy.ndarray`
        A rotation matrix which transforms `vector1` to `vector2`.

    References
    ----------
    http://tinyurl.com/kybj9ox
    http://tinyurl.com/gn6e8mz

    """

    # Make sure both inputs are unit vectors.
    vector1 = normalize_vector(vector1)
    vector2 = normalize_vector(vector2)

    # Hande the case where the input and output vectors are equal.
    if np.allclose(vector1, vector2, atol=1e-8):
        return np.identity(3)

    # Handle the case where the rotation is 180 degrees.
    if np.allclose(vector1, np.multiply(vector2, -1), atol=1e-8):
        return rotation_matrix_arbitrary_axis(
            angle=np.pi, axis=orthogonal_vector(vector1)
        )

    v = np.cross(vector1, vector2)
    vx = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    s = np.linalg.norm(v)
    c = np.dot(vector1, vector2)
    i = np.identity(3)
    mult_factor = (1 - c) / np.square(s)

    # Initialize as a scipy Rotation object, which normalizes the
    # matrix and allows for returns as quaternion or alternative
    # type in the future.
    return Rotation.from_matrix(
        i + vx + np.multiply(np.dot(vx, vx), mult_factor)
    ).as_matrix()


def rotation_matrix_arbitrary_axis(angle, axis):
    """
    Returns a rotation matrix of `angle` radians about `axis`.

    Parameters
    ----------
    angle : :class:`float`
        The size of the rotation in radians.

    axis : :class:`numpy.ndarray`
        A 3 element aray which represents a vector. The vector is the
        axis about which the rotation is carried out. Must be of
        unit magnitude.

    Returns
    -------
    :class:`numpy.ndarray`
        A ``3x3`` array representing a rotation matrix.

    """

    a = np.cos(angle / 2)
    b, c, d = axis * np.sin(angle / 2)

    e11 = np.square(a) + np.square(b) - np.square(c) - np.square(d)
    e12 = 2 * (b * c - a * d)
    e13 = 2 * (b * d + a * c)

    e21 = 2 * (b * c + a * d)
    e22 = np.square(a) + np.square(c) - np.square(b) - np.square(d)
    e23 = 2 * (c * d - a * b)

    e31 = 2 * (b * d - a * c)
    e32 = 2 * (c * d + a * b)
    e33 = np.square(a) + np.square(d) - np.square(b) - np.square(c)

    # Initialize as a scipy Rotation object, which normalizes the
    # matrix and allows for returns as quaternion or alternative
    # type in the future.
    return Rotation.from_matrix(
        np.array([[e11, e12, e13], [e21, e22, e23], [e31, e32, e33]])
    ).as_matrix()


def dice_similarity(mol1, mol2, fp_radius=3):
    """
    Return the chemical similarity between two molecules.

    Parameters
    ----------
    mol1 : :class:`.Molecule`
        The first molecule.

    mol2 : :class:`.Molecule`
        The second molecule.

    fp_radius : :class:`int`, optional
        The radius of the Morgan fingerprint used to calculate
        similarity.

    Returns
    -------
    :class:`float`
        The similarity.

    """

    rdkit_mol1 = mol1.to_rdkit_mol()
    rdkit.SanitizeMol(rdkit_mol1)
    fp1 = rdkit.GetMorganFingerprint(
        mol=rdkit_mol1,
        radius=fp_radius,
    )
    rdkit_mol2 = mol2.to_rdkit_mol()
    rdkit.SanitizeMol(rdkit_mol2)
    fp2 = rdkit.GetMorganFingerprint(
        mol=rdkit_mol2,
        radius=fp_radius,
    )
    return rdkit.DataStructs.DiceSimilarity(fp1, fp2)


def quaternion(u):
    """
    Returns a translation + rotation quaternion.

    Parameters
    ----------
    u : :class:`list` of :class:`float`
        A :class:`list` of length 3 holding the parameter for the
        quarternion.

    References
    ----------
    K. Shoemake, Uniform random rotations, Graphics Gems III,
    pages 124-132. Academic, New York, 1992.

    """

    a, b, c = u
    q = np.zeros(4, np.float64)
    q[0] = np.sqrt(1.0 - a) * np.sin(2.0 * np.pi * b)
    q[1] = np.sqrt(1.0 - a) * np.cos(2.0 * np.pi * b)
    q[2] = np.sqrt(a) * np.sin(2.0 * np.pi * c)
    q[3] = np.sqrt(a) * np.cos(2.0 * np.pi * c)
    return q


def translation_component(q):
    """
    Extracts translation vector from quaternion.

    Parameters
    ----------
    q : :class:`numpy.ndarray`
        A length 4 quaternion.

    Returns
    -------
    :class:`numpy.ndarray`
        The translation vector encoded within `q`.

    """

    rot_epsilon = 1e-6
    q = np.copy(q)
    if q[0] < 0.0:
        q = -q
    if q[0] > 1.0:
        q /= np.sqrt(np.dot(q, q))

    theta = 2.0 * np.arccos(q[0])
    s = np.sqrt(1.0 - q[0] * q[0])
    if s < rot_epsilon:
        p = 2.0 * q[1:4]
    else:
        p = q[1:4] / s * theta
    return p


def tar_output():
    """
    Places all the content in the ``output`` folder into a .tgz file.

    Returns
    -------
    None : :class:`NoneType`

    """

    tname = os.path.join("output", "output.tgz")
    with tarfile.open(tname, "w:gz") as tar:
        tar.add("output")


def _printer(time_taken):
    m, s = divmod(time_taken, 60)
    h, m = divmod(m, 60)
    print(f"\nTime taken was {int(h)} : {int(m)} : {int(s)}.\n\n")


@contextmanager
def time_it(output=_printer):
    """
    Times the code executed within the indent.

    Parameters
    ----------
    output : :class:`callable`, optional
        A function which takes the time taken for the code block to
        execute as its only input. Can be used to redirect or format
        the output.

    Examples
    --------
    *Timing the Execution of a Function*

    This is a context manager so it should be used as:

    .. testsetup:: timing-the-execution-of-a-function

        import stk
        # Do not print the output in the test because the time taken
        # is expected to vary wildy.
        _time_it = stk.time_it
        stk.time_it = lambda: _time_it(
            output=lambda time_taken: print('test'),
        )

    .. testcode:: timing-the-execution-of-a-function

        import stk

        with stk.time_it():
            print('a')
            print('b')
            print('c')

    .. testoutput:: timing-the-execution-of-a-function
        :hide:

        a
        b
        c
        test

    .. testcleanup:: timing-the-execution-of-a-function

        stk.time_it = _time_it

    After all 3 functions are finished and the nested block is exited
    the time taken to process the entire block is printed.

    """

    start = time.time()
    yield
    output(time.time() - start)


def vector_angle(vector1, vector2):
    """
    Returns the angle between two vectors in radians.

    Parameters
    ----------
    vector1 : :class:`numpy.ndarray`
        The first vector.

    vector2 : :class:`numpy.ndarray`
        The second vector.

    Returns
    -------
    :class:`float`
        The angle between `vector1` and `vector2` in radians.

    """

    if np.all(np.equal(vector1, vector2)):
        return 0.0

    numerator = np.dot(vector1, vector2)
    denominator = np.linalg.norm(vector1) * np.linalg.norm(vector2)
    # This if statement prevents returns of NaN due to floating point
    # inaccuracy.
    term = numerator / denominator
    if term >= 1.0:
        return 0.0
    if term <= -1.0:
        return np.pi
    return np.arccos(term)


def get_acute_vector(reference, vector):
    if (
        # vector_angle is NaN if reference is [0, 0, 0].
        not np.allclose(reference, [0, 0, 0], atol=1e-5)
        and vector_angle(vector, reference) > np.pi / 2
    ):
        return vector * -1
    return vector


def get_plane_normal(points):
    centroid = points.sum(axis=0) / len(points)
    return np.around(np.linalg.svd(points - centroid)[-1][2, :], 14)


def cap_absolute_value(value, max_absolute_value=1):
    """
    Returns `value` with absolute value capped at `max_absolute_value`.

    Particularly useful in passing values to trignometric functions
    where numerical errors may result in an argument > 1 being passed
    in.

    This code is modified from the pymatgen source code [1]_.

    Parameters
    ----------
    value : :class:`float`
        Value to cap.

    max_absolute_value : :class:`float`, optional
        Absolute value to cap `value` at.
        Defaults to 1.

    Returns
    -------
    :class:`float`
        `value` capped at `max_absolute_value` with sign preserved.

    References
    ----------
    .. [1] https://pymatgen.org/pymatgen.util.num.html

    """

    return max(min(value, max_absolute_value), -max_absolute_value)



from .alcohol import *  # noqa
from .aldehyde import *  # noqa
from .alkene import *  # noqa
from .alkyne import *  # noqa
from .amide import *  # noqa
from .boronic_acid import *  # noqa
from .bromo import *  # noqa
from .carboxylic_acid import *  # noqa
from .dibromo import *  # noqa
from .difluoro import *  # noqa
from .diol import *  # noqa
from .fluoro import *  # noqa
from .iodo import *  # noqa
from .primary_amino import *  # noqa
from .ring_amine import *  # noqa
from .secondary_amino import *  # noqa
from .single_atom import *  # noqa
from .thioacid import *  # noqa
from .thiol import *  # noqa



from .utilities import is_equivalent_atom


def test_clone(atom):
    """
    Test :meth:`.Atom.clone`.

    Parameters
    ----------
    atom : :class:`.Atom`
        The atom to be cloned.

    Returns
    -------
    None : :class:`NoneType`

    """

    clone = atom.clone()
    is_equivalent_atom(atom, clone)



import typing
from collections.abc import Callable, Iterator, Sequence

import numpy as np

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.selection.batch import Batch
from stk._internal.ea.selection.selectors.yielded_batches import YieldedBatches
from stk._internal.key_makers.inchi import Inchi
from stk._internal.key_makers.molecule import MoleculeKeyMaker

from .selector import Selector

T = typing.TypeVar("T", bound=MoleculeRecord)


class Tournament(Selector[T]):
    """
    Yields batches of molecules through tournament selection.

    In tournament selection, a random number of batches is chosen from
    the population to undergo a competition. In each competition, the
    batch with the highest fitness value is yielded. This is repeated
    until `num_batches` are yielded.

    Examples:

        *Yielding Single Molecule Batches*

        Yielding molecules one at a time. For example, if molecules need
        to be selected for mutation or the next generation.

        .. testcode:: yielding-single-molecule-batches

            import stk

            # Make the selector.
            tournament = stk.Tournament(
                num_batches=5,
                batch_size=1
            )

            population = {
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=[
                            stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                        ],
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): i
                for i in range(100)
            }

            # Select the molecules.
            for selected, in tournament.select(population):
                # Do stuff with each selected molecule.
                pass
    """

    def __init__(
        self,
        num_batches: int | None = None,
        batch_size: int = 1,
        duplicate_molecules: bool = True,
        duplicate_batches: bool = True,
        key_maker: MoleculeKeyMaker = Inchi(),
        fitness_modifier: Callable[
            [dict[T, float]], dict[T, float]
        ] = lambda x: x,
        random_seed: int | np.random.Generator | None = None,
    ) -> None:
        """
        Parameters:
            num_batches:
                The number of batches to yield. If ``None`` then yielding
                will continue forever or until the generator is exhausted,
                whichever comes first.

            batch_size:
                The number of molecules yielded at once.

            duplicate_molecules:
                If ``True`` the same molecule can be yielded in more than
                one batch.

            duplicate_batches:
                If ``True`` the same batch can be yielded more than once.

            key_maker:
                Used to get the keys of molecules. If two molecules have
                the same key, they are considered duplicates.

            fitness_modifier:
                Takes the `population` on which :meth:`~.Selector.select`
                is called and returns a :class:`dict`, which maps records
                in the `population` to the fitness values the
                :class:`.Selector` should use.

            random_seed:
                The random seed to use.
        """
        super().__init__(key_maker, fitness_modifier, batch_size)

        if random_seed is None or isinstance(random_seed, int):
            random_seed = np.random.default_rng(random_seed)

        self._generator = random_seed
        self._duplicate_molecules = duplicate_molecules
        self._duplicate_batches = duplicate_batches
        self._num_batches = (
            float("inf") if num_batches is None else num_batches
        )

    def _select_from_batches(
        self,
        batches: Sequence[Batch[T]],
        yielded_batches: YieldedBatches[T],
    ) -> Iterator[Batch[T]]:
        # The tournament can only take place if there is more than 1
        # batch.
        while (
            len(batches) > 1 and yielded_batches.get_num() < self._num_batches
        ):
            tournament_size = self._generator.integers(
                low=2, high=len(batches) + 1
            )
            competitors = self._generator.choice(
                a=batches,  # type: ignore
                size=tournament_size,
                replace=False,
            )
            yield max(competitors)

            if not self._duplicate_molecules:
                batches_ = filter(
                    yielded_batches.has_no_yielded_molecules,
                    batches,
                )
            if not self._duplicate_batches:
                batches_ = filter(
                    yielded_batches.is_unyielded_batch,
                    batches,
                )
            if not self._duplicate_molecules or not self._duplicate_batches:
                batches = tuple(batches_)



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def thioacid(get_atom_ids):
    a, b, c, d, e = get_atom_ids(5)
    return _thioacid(stk.C(a), stk.O(b), stk.S(c), stk.H(d), stk.C(e))


def _thioacid(carbon, oxygen, sulfur, hydrogen, atom):
    bonders = ()
    deleters = ()
    return GenericCaseData(
        functional_group=stk.Thioacid(
            carbon=carbon,
            oxygen=oxygen,
            sulfur=sulfur,
            hydrogen=hydrogen,
            atom=atom,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(carbon, oxygen, sulfur, hydrogen, atom),
        bonders=bonders,
        deleters=deleters,
    )



from stk._internal.topology_graphs.cof.cof import Cof
from stk._internal.topology_graphs.cof.edge import CofEdge as Edge
from stk._internal.topology_graphs.cof.hexagonal import Hexagonal
from stk._internal.topology_graphs.cof.honeycomb import Honeycomb
from stk._internal.topology_graphs.cof.kagome import Kagome
from stk._internal.topology_graphs.cof.linkerless_honeycomb import (
    LinkerlessHoneycomb,
)
from stk._internal.topology_graphs.cof.periodic_hexagonal import (
    PeriodicHexagonal,
)
from stk._internal.topology_graphs.cof.periodic_honeycomb import (
    PeriodicHoneycomb,
)
from stk._internal.topology_graphs.cof.periodic_kagome import PeriodicKagome
from stk._internal.topology_graphs.cof.periodic_linkerless_honeycomb import (
    PeriodicLinkerlessHoneycomb,
)
from stk._internal.topology_graphs.cof.periodic_square import PeriodicSquare
from stk._internal.topology_graphs.cof.square import Square
from stk._internal.topology_graphs.cof.vertices import (
    LinearVertex,
    NonLinearVertex,
    UnaligningVertex,
)

__all__ = [
    "LinearVertex",
    "NonLinearVertex",
    "UnaligningVertex",
    "Cof",
    "Hexagonal",
    "Honeycomb",
    "Kagome",
    "Square",
    "Edge",
    "PeriodicHoneycomb",
    "PeriodicSquare",
    "PeriodicKagome",
    "PeriodicHexagonal",
    "PeriodicLinkerlessHoneycomb",
    "LinkerlessHoneycomb",
]



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def boronic_acid(get_atom_ids):
    a, b, c, d, e, f = get_atom_ids(6)
    return _boronic_acid(
        boron=stk.B(a),
        oxygen1=stk.O(b),
        hydrogen1=stk.H(c),
        oxygen2=stk.O(d),
        hydrogen2=stk.H(e),
        atom=stk.C(f),
    )


def _boronic_acid(boron, oxygen1, hydrogen1, oxygen2, hydrogen2, atom):
    bonders = (oxygen1, oxygen2)
    deleters = (hydrogen1, hydrogen2)
    return GenericCaseData(
        functional_group=stk.BoronicAcid(
            boron=boron,
            oxygen1=oxygen1,
            hydrogen1=hydrogen1,
            oxygen2=oxygen2,
            hydrogen2=hydrogen2,
            atom=atom,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(boron, oxygen1, hydrogen1, oxygen2, hydrogen2, atom),
        bonders=bonders,
        deleters=deleters,
    )



import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda: CaseData(
            key_maker=stk.Inchi(),
            molecule=stk.BuildingBlock("NCCN"),
            key_name="InChI",
            key="InChI=1S/C2H8N2/c3-1-2-4/h1-4H2",
        ),
        lambda: CaseData(
            key_maker=stk.InchiKey(),
            molecule=stk.BuildingBlock("NCCN"),
            key_name="InChIKey",
            key="PIICEJLVQHRZGT-UHFFFAOYSA-N",
        ),
        lambda: CaseData(
            key_maker=stk.Smiles(),
            molecule=stk.BuildingBlock("NCCN"),
            key_name="SMILES",
            key="NCCN",
        ),
        lambda: CaseData(
            key_maker=stk.Smiles(),
            molecule=stk.BuildingBlock("C(N)CN"),
            key_name="SMILES",
            key="NCCN",
        ),
        lambda: CaseData(
            key_maker=stk.Smiles(),
            molecule=stk.BuildingBlock(
                "C(#Cc1cccc2cnccc12)c1ccc2[nH]c3ccc(C#Cc4cccc5ccncc45)cc3c2c1"
            ),
            key_name="SMILES",
            key=(
                "C(#Cc1cccc2cnccc12)c1ccc2[nH]c3ccc(C#Cc4cccc5ccncc45)cc3c2c1"
            ),
        ),
        lambda: CaseData(
            key_maker=stk.Smiles(),
            molecule=stk.BuildingBlock(
                "C(#Cc1cccc2cnccc21)c1ccc2[nH]c3ccc(C#Cc4cccc5ccncc54)cc3c2c1"
            ),
            key_name="SMILES",
            key=(
                "C(#Cc1cccc2cnccc12)c1ccc2[nH]c3ccc(C#Cc4cccc5ccncc45)cc3c2c1"
            ),
        ),
        lambda: CaseData(
            key_maker=stk.Smiles(),
            molecule=stk.BuildingBlock("C[C@H](O)c1ccccc1"),
            key_name="SMILES",
            key="C[C@H](O)c1ccccc1",
        ),
        lambda: CaseData(
            key_maker=stk.MoleculeKeyMaker(
                key_name="NumAtoms",
                get_key=lambda molecule: molecule.get_num_atoms(),
            ),
            molecule=stk.BuildingBlock("NCCN"),
            key_name="NumAtoms",
            key=12,
        ),
    ),
)
def case_data(request) -> CaseData:
    return request.param()



from __future__ import annotations

import stk

from tests.utilities import is_equivalent_constructed_molecule

from .case_data import CaseData


def test_get_all(case_data: CaseData) -> None:
    """
    Test iteration over all entries.

    Parameters:

        case_data: A test case. Holds the databases to test and the
            molecules to place into the databases.

    """

    _test_get_all(
        database=case_data.database,
        expected_molecules=case_data.expected_molecules,
    )


def _test_get_all(
    database: stk.ConstructedMoleculeDatabase,
    expected_molecules: dict[str, stk.ConstructedMolecule],
) -> None:
    """
    Test iteration over all entries.

    Parameters:

        database: A database to test.

        expected_molecules: The expected molecules to get from the
            databases using their smiles as the key.

    """

    smiles = stk.Smiles()
    for i, retrieved in enumerate(database.get_all()):
        expected = expected_molecules[smiles.get_key(retrieved)]
        is_equivalent_constructed_molecule(
            constructed_molecule1=(expected.with_canonical_atom_ordering()),
            constructed_molecule2=(retrieved.with_canonical_atom_ordering()),
        )

    assert i + 1 == len(expected_molecules)



from tests.utilities import is_equivalent_constructed_molecule


def is_clone_constructed_molecule(
    constructed_molecule1,
    constructed_molecule2,
):
    assert constructed_molecule1 is not constructed_molecule2
    is_equivalent_constructed_molecule(
        constructed_molecule1=constructed_molecule1,
        constructed_molecule2=constructed_molecule2,
    )



import numpy as np
import pytest

from ..utilities import get_direction


@pytest.mark.parametrize(
    argnames=("position_matrix", "atom_ids", "direction"),
    argvalues=(
        (
            np.array(
                [
                    [1.0, 0.0, 0.0],
                    [2.0, 0.0, 0.0],
                    [3.0, 0.0, 0.0],
                ]
            ),
            (0, 1, 2),
            np.array([1.0, 0.0, 0.0]),
        ),
        (
            np.array(
                [
                    [0.0, 1.0, 0.0],
                    [0.0, 2.0, 0.0],
                    [0.0, 3.0, 0.0],
                ]
            ),
            (0, 1, 2),
            np.array([0.0, 1.0, 0.0]),
        ),
        (
            np.array(
                [
                    [1.0, 0.0, 0.0],
                    [2.0, 0.0, 0.0],
                    [0.0, 5.0, 0.0],
                ]
            ),
            (0, 1),
            np.array([1.0, 0.0, 0.0]),
        ),
    ),
)
def test_get_direction_helper(position_matrix, atom_ids, direction):
    """
    Test the :func:`.get_direction`.

    """

    # The direction can be either parallel or anti-parallel.
    result = get_direction(position_matrix, atom_ids)
    assert abs(abs(result @ direction) - 1) < 1e-14



import itertools as it

from ...utilities import is_clone_generic_functional_group


def test_with_atoms(generic_functional_group, get_atom_map):
    """
    Test :meth:`.GenericFunctionalGroup.with_atoms`.

    Parameters
    ----------
    generic_functional_group : :class:`.GenericFunctionalGroup`
        The functional group to test.

    get_atom_map : :class:`callable`
        Takes a single parameter, `functional_group`, and returns a
        valid `atom_map` parameter for its
        :meth:`.GenericFunctionalGroup.with_atoms` method.
        This allows the testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Save a clone to ensure that "generic_functional_group" is not
    # changed by the test, because it meant to be immutable.
    before = generic_functional_group.clone()
    _test_with_atoms(generic_functional_group, get_atom_map)
    is_clone_generic_functional_group(before, generic_functional_group)


def _test_with_atoms(functional_group, get_atom_map):
    """
    Test :meth:`.GenericFunctionalGroup.with_atoms`.

    Parameters
    ----------
    generic_functional_group : :class:`.GenericFunctionalGroup`
        The functional group to test.

    get_atom_map : :class:`callable`
        Takes a single parameter, `functional_group`, and returns a
        valid `atom_map` parameter for its
        :meth:`.GenericFunctionalGroup.with_atoms` method.
        This allows the testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    atom_map = get_atom_map(functional_group)
    clone = functional_group.with_atoms(atom_map)

    is_modified_sequence(
        atoms1=functional_group.get_atoms(),
        atoms2=clone.get_atoms(),
        atom_map=atom_map,
    )
    is_modified_id_sequence(
        ids1=functional_group.get_atom_ids(),
        ids2=clone.get_atom_ids(),
        atom_map=atom_map,
    )

    is_modified_sequence(
        atoms1=functional_group.get_bonders(),
        atoms2=clone.get_bonders(),
        atom_map=atom_map,
    )
    is_modified_id_sequence(
        ids1=functional_group.get_bonder_ids(),
        ids2=clone.get_bonder_ids(),
        atom_map=atom_map,
    )

    is_modified_sequence(
        atoms1=functional_group.get_deleters(),
        atoms2=clone.get_deleters(),
        atom_map=atom_map,
    )
    is_modified_id_sequence(
        ids1=functional_group.get_deleter_ids(),
        ids2=clone.get_deleter_ids(),
        atom_map=atom_map,
    )

    is_modified_id_sequence(
        ids1=functional_group.get_placer_ids(),
        ids2=clone.get_placer_ids(),
        atom_map=atom_map,
    )
    is_modified_id_sequence(
        ids1=functional_group.get_core_atom_ids(),
        ids2=clone.get_core_atom_ids(),
        atom_map=atom_map,
    )


def is_modified_sequence(atoms1, atoms2, atom_map):
    for atom1, atom2 in it.zip_longest(atoms1, atoms2):
        assert atom2 is atom_map.get(atom1.get_id(), atom1)


def is_modified_id_sequence(ids1, ids2, atom_map):
    for id1, id2 in it.zip_longest(ids1, ids2):
        if id1 in atom_map:
            assert id2 == atom_map[id1].get_id()
        else:
            assert id2 == id1



import itertools as it


def test_get_deleter_ids(generic_case_data):
    """
    Test :meth:`.GenericFunctionalGroup.get_deleter_ids`.

    Parameters
    ----------
    generic_case_data : :class:`.GenericCaseData`
        The test case. Holds the functional group to test and the
        atoms holding the correct deleter ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_deleter_ids(
        functional_group=generic_case_data.functional_group,
        deleters=generic_case_data.deleters,
    )


def _test_get_deleter_ids(functional_group, deleters):
    """
    Test :meth:`.GenericFunctionalGroup.get_deleter_ids`.

    Parameters
    ----------
    functional_group : :class:`.GenericFunctionalGroup`
        The functional group to test.

    deleters : :class:`tuple` of :class:`.Atom`
        The atoms holding the correct deleter ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    for id_, atom in it.zip_longest(
        functional_group.get_deleter_ids(),
        deleters,
    ):
        assert id_ == atom.get_id()



import pymongo
import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--mongodb_uri",
        action="store",
        default="mongodb://localhost:27017/",
    )


@pytest.fixture(
    scope="session",
)
def mongo_client(pytestconfig):
    return pymongo.MongoClient(pytestconfig.getoption("mongodb_uri"))



import argparse
import logging
import pathlib
import shutil

import numpy as np
import pytest

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s | %(levelname)s | %(message)s",
)
logger = logging.getLogger(__name__)


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "test_path",
        type=pathlib.Path,
        help="where are the tests?",
    )
    parser.add_argument(
        "output_path",
        type=pathlib.Path,
        help="where to save structures to",
    )

    return parser.parse_args()


def main() -> None:
    """Run script."""
    args = _parse_args()
    tests = args.test_path

    pos_mats = (
        tests / "molecular/molecules/molecule/fixtures/position_matrices/"
    )
    old_pos_mats = (
        tests / "molecular/molecules/molecule/fixtures/old_position_matrices/"
    )
    args.output_path.mkdir(parents=True, exist_ok=True)

    # Copy from current to old and remove all in current while maintaining
    # directory for tests.
    pos_mats.rename(old_pos_mats)
    pos_mats.mkdir(parents=True, exist_ok=False)

    # Run pytest, but only those that will make the position matrices.
    pytest.main(["-k test_get_plane_normal"])

    # Compare.
    new_numpy_files = list(sorted(pos_mats.glob("*.npy")))
    old_numpy_files = list(sorted(old_pos_mats.glob("*.npy")))
    assert len(new_numpy_files) == len(old_numpy_files)

    rmsds = []
    for new_file, old_file in zip(
        new_numpy_files,
        old_numpy_files,
        strict=True,
    ):
        assert new_file.stem == old_file.stem
        new_array = np.load(new_file)
        old_array = np.load(old_file)
        assert len(new_array) == len(old_array)

        rmsd = np.sqrt(np.mean((new_array - old_array) ** 2))

        rmsds.append(rmsd)
        if rmsd >= 0.1:
            logger.info(
                "very large rmsd for %s is %s", old_file.name, round(rmsd, 3)
            )
            output_file = args.output_path / f"{old_file.stem}_0.xyz"
            top_line = f"{len(old_array)}\n"
            comment_line = "\n"
            pos_lines = []
            for pos in old_array:
                pos_lines.append(f"C {pos[0]:.6f} {pos[1]:.6f} {pos[2]:.6f}\n")
            with output_file.open("w") as f:
                f.write(top_line)
                f.write(comment_line)
                f.writelines(pos_lines)

            new_output_file = args.output_path / f"{new_file.stem}_1.xyz"
            top_line = f"{len(new_array)}\n"
            comment_line = "\n"
            pos_lines = []
            for pos in new_array:
                pos_lines.append(f"C {pos[0]:.6f} {pos[1]:.6f} {pos[2]:.6f}\n")
            with new_output_file.open("w") as f:
                f.write(top_line)
                f.write(comment_line)
                f.writelines(pos_lines)

    logger.info("average rmsd is %s", round(np.mean(rmsds), 3))

    # Delete test and return back to normal.
    shutil.rmtree(pos_mats)
    old_pos_mats.rename(pos_mats)
    logger.info("do not forget to delete %s", args.output_path)


if __name__ == "__main__":
    main()



import stk

from .constructed_molecule import is_equivalent_constructed_molecule
from .molecule import is_equivalent_molecule


def is_equivalent(molecule1, molecule2):
    is_constructed_molecule1 = isinstance(
        molecule1,
        stk.ConstructedMolecule,
    )
    is_constructed_molecule2 = isinstance(
        molecule2,
        stk.ConstructedMolecule,
    )
    if is_constructed_molecule1 and is_constructed_molecule2:
        is_equivalent_constructed_molecule(molecule1, molecule2)
        return

    is_equivalent_molecule(molecule1, molecule2)



"""
Four Plus Eight
===============

"""

import numpy as np

from stk._internal.topology_graphs.cage.cage import Cage
from stk._internal.topology_graphs.cage.vertices import (
    LinearVertex,
    NonLinearVertex,
)
from stk._internal.topology_graphs.edge import Edge


class FourPlusEight(Cage):
    """
    Represents a cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='C1=CC=C(C(=C1)Br)Br',
            functional_groups=[stk.BromoFactory()],
        )
        bb2 = stk.BuildingBlock(
            smiles='C1(=C(C(=C1Br)Br)Br)Br',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.FourPlusEight((bb1, bb2)),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.Collapser` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='C1=CC=C(C(=C1)Br)Br',
            functional_groups=[stk.BromoFactory()],
        )
        bb2 = stk.BuildingBlock(
            smiles='C1(=C(C(=C1Br)Br)Br)Br',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.FourPlusEight(
                building_blocks=(bb1, bb2),
                optimizer=stk.Collapser(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    Nonlinear building blocks with four functional groups are
    required for this topology.

    Linear building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 4-functional groups: 0 to 3
        | 2-functional groups: 4 to 11

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        NonLinearVertex(0, np.array([-1, -1, 0]), False),
        NonLinearVertex(1, np.array([-1, 1, 0]), False),
        NonLinearVertex(2, np.array([1, -1, 0]), False),
        NonLinearVertex(3, np.array([1, 1, 0]), False),
        LinearVertex(4, np.array([-2, 0, 1]), False),
        LinearVertex(5, np.array([-2, 0, -1]), False),
        LinearVertex(6, np.array([0, 2, 1]), False),
        LinearVertex(7, np.array([0, 2, -1]), False),
        LinearVertex(8, np.array([0, -2, 1]), False),
        LinearVertex(9, np.array([0, -2, -1]), False),
        LinearVertex(10, np.array([2, 0, 1]), False),
        LinearVertex(11, np.array([2, 0, -1]), False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[4], _vertex_prototypes[1]),
        Edge(2, _vertex_prototypes[5], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[5], _vertex_prototypes[1]),
        Edge(4, _vertex_prototypes[6], _vertex_prototypes[1]),
        Edge(5, _vertex_prototypes[6], _vertex_prototypes[3]),
        Edge(6, _vertex_prototypes[7], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[7], _vertex_prototypes[3]),
        Edge(8, _vertex_prototypes[8], _vertex_prototypes[0]),
        Edge(9, _vertex_prototypes[8], _vertex_prototypes[2]),
        Edge(10, _vertex_prototypes[9], _vertex_prototypes[0]),
        Edge(11, _vertex_prototypes[9], _vertex_prototypes[2]),
        Edge(12, _vertex_prototypes[10], _vertex_prototypes[2]),
        Edge(13, _vertex_prototypes[10], _vertex_prototypes[3]),
        Edge(14, _vertex_prototypes[11], _vertex_prototypes[2]),
        Edge(15, _vertex_prototypes[11], _vertex_prototypes[3]),
    )

    _num_windows = 6
    _num_window_types = 2



import numpy as np
import pytest
import stk
from rdkit.Chem import AllChem as rdkit

from ....case_data import CaseData

single_atom = rdkit.MolFromSmiles("[Fe]")
single_atom.AddConformer(rdkit.Conformer(single_atom.GetNumAtoms()))


@pytest.fixture(
    params=(
        lambda: CaseData(
            vertex=stk.metal_complex.MetalVertex(
                id=0,
                position=(1, 2, 3),
            ),
            edges=(),
            building_block=stk.BuildingBlock.init_from_rdkit_mol(single_atom),
            position=np.array([1, 2, 3], dtype=np.float64),
            alignment_tests={},
            functional_group_edges={},
        ),
    ),
)
def metal(request) -> CaseData:
    return request.param()



"""
XYZ Writer
==========

"""

from __future__ import annotations

import pathlib
import typing

from stk._internal.molecule import Molecule
from stk._internal.utilities.utilities import OneOrMany


class XyzWriter:
    """
    A writer class for ``.xyz`` files.

    Examples:

        *Writing to a File*

        .. testcode:: writing-to-a-file

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])

            writer = stk.XyzWriter()
            writer.write(molecule=bb1, path='bb1.xyz')

        .. testcode:: writing-to-a-file
            :hide:

            import os

            assert os.path.exists('bb1.xyz')

        .. testcleanup:: writing-to-a-file

            os.remove('bb1.xyz')

    """

    def _write_content(
        self,
        molecule: Molecule,
        atom_ids: typing.Optional[OneOrMany[int]],
    ) -> list[str]:
        if atom_ids is None:
            atom_ids = range(molecule.get_num_atoms())
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)

        coords = molecule.get_position_matrix()
        content = [""]
        for i, atom_id in enumerate(atom_ids, 1):
            x, y, z = (i for i in coords[atom_id])
            (atom,) = molecule.get_atoms(atom_ids=atom_id)
            symbol = atom.__class__.__name__
            content.append(f"{symbol} {x:f} {y:f} {z:f}\n")
        # Set first line to the atom_count.
        content[0] = f"{i}\n\n"

        return content

    def to_string(
        self,
        molecule: Molecule,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
    ) -> str:
        """
        Get the ``.xyz`` string of  `molecule`.

        Parameters:

            molecule:
                Molecule to write to `.xyz` format.

            atom_ids:
                The atom ids of atoms to write. Can be a single
                :class:`int`, if a single atom is to be used, or
                ``None``, if all atoms are to be used.

        Returns:

            String in ``.xyz`` file format.

        """

        content = self._write_content(molecule, atom_ids)

        return "".join(content)

    def write(
        self,
        molecule: Molecule,
        path: pathlib.Path | str,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
    ) -> None:
        """
        Write `molecule` to ``.xyz`` file format.

        Parameters:

            molecule:
                Molecule to write to ``.xyz`` format.

            path:
                The full path to the file being written.

            atom_ids:
                The atom ids of atoms to write. Can be a single
                :class:`int`, if a single atom is to be used, or
                ``None``, if all atoms are to be used.

        """

        content = self._write_content(molecule, atom_ids)

        with open(path, "w") as f:
            f.write("".join(content))



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def amide(get_atom_ids):
    a, b, c, d, e, f = get_atom_ids(6)
    return _amide(
        carbon=stk.C(a),
        oxygen=stk.O(b),
        nitrogen=stk.N(c),
        hydrogen1=stk.H(d),
        hydrogen2=stk.H(e),
        atom=stk.C(f),
    )


def _amide(carbon, oxygen, nitrogen, hydrogen1, hydrogen2, atom):
    bonders = (carbon,)
    deleters = (nitrogen, hydrogen1, hydrogen2)
    return GenericCaseData(
        functional_group=stk.Amide(
            carbon=carbon,
            oxygen=oxygen,
            nitrogen=nitrogen,
            hydrogen1=hydrogen1,
            hydrogen2=hydrogen2,
            atom=atom,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(carbon, oxygen, nitrogen, hydrogen1, hydrogen2, atom),
        bonders=bonders,
        deleters=deleters,
    )



class CaseData:
    """
    A test case.

    Attributes
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    num_atoms : :class:`int`
        The correct number of atoms :attr:`.molecule` should have.

    """

    def __init__(self, molecule, num_atoms):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule to test.

        num_atoms : :class:`int`
            The correct number of atoms `molecule` should have.

        """

        self.molecule = molecule
        self.num_atoms = num_atoms



from typing import Any

import pandas as pd
import pytest
import stk

from .case_data import CaseData


def _get_topology_graph() -> stk.polymer.Linear:
    return stk.polymer.Linear(
        building_blocks=[
            stk.BuildingBlock("BrCCBr", stk.BromoFactory()),
        ],
        repeating_unit="A",
        num_repeating_units=2,
    )


def get_generation(
    *fitness_values: float,
) -> dict[stk.MoleculeRecord[Any], float | None]:
    v1, v2, v3, *_ = fitness_values
    topology_graph = _get_topology_graph()
    return {
        stk.MoleculeRecord(
            topology_graph=topology_graph,
        ): v1,
        stk.MoleculeRecord(
            topology_graph=topology_graph,
        ): v2,
        stk.MoleculeRecord(
            topology_graph=topology_graph,
        ): v3,
        stk.MoleculeRecord(
            topology_graph=topology_graph,
        ): None,
    }


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            plotter=stk.ProgressPlotter(
                property=[
                    (0, 1, 2),
                    (10, 20, 30),
                    (40, 50, 60),
                    (40, 50, 60),
                    (70, 80, 90),
                ],
                y_label="Fitness Value",
            ),
            plot_data=pd.DataFrame(
                {
                    "Generation": [0] * 3
                    + [1] * 3
                    + [2] * 3
                    + [3] * 3
                    + [4] * 3,
                    "Fitness Value": [
                        2.0,
                        1.0,
                        0.0,
                        30.0,
                        20.0,
                        10.0,
                        60.0,
                        50.0,
                        40.0,
                        60.0,
                        50.0,
                        40.0,
                        90.0,
                        80.0,
                        70.0,
                    ],
                    "Type": ["Max", "Mean", "Min"] * 5,
                }
            ),
        ),
    ),
)
def case_data(request: pytest.FixtureRequest) -> CaseData:
    return request.param()



"""
Thiol
=====

"""

from .generic_functional_group import GenericFunctionalGroup


class Thiol(GenericFunctionalGroup):
    """
    Represents a thiol functional group.


    The structure of the functional group is given by the pseudo-SMILES
    ``[atom][sulfur][hydrogen]``.

    """

    def __init__(
        self,
        sulfur,
        hydrogen,
        atom,
        bonders,
        deleters,
        placers=None,
    ):
        """
        Initialize a :class:`.Thiol` instance.

        Parameters
        ----------
        sulfur : :class:`.S`
            The ``[sulfur]`` atom.

        hydrogen : :class:`.H`
            The ``[hydrogen]`` atom.

        atom : :class:`.Atom`
            The ``[atom]`` atom.

        bonders : :class:`tuple` of :class:`.Atom`
            The bonder atoms.

        deleters : :class:`tuple` of :class:`.Atom`
            The deleter atoms.

        placers : :class:`tuple` of :class:`.Atom`, optional
            The placer atoms. If ``None`` the `bonders` will be used.

        """

        self._sulfur = sulfur
        self._hydrogen = hydrogen
        self._atom = atom
        super().__init__(
            atoms=(sulfur, hydrogen, atom),
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_sulfur(self):
        """
        Get the ``[sulfur]`` atom.

        Returns
        -------
        :class:`.S`
            The ``[sulfur]`` atom.

        """

        return self._sulfur

    def get_hydrogen(self):
        """
        Get the ``[hydrogen]`` atom.

        Returns
        -------
        :class:`.H`
            The ``[hydrogen]`` atom.

        """

        return self._hydrogen

    def get_atom(self):
        """
        Get the ``[atom]`` atom.

        Returns
        -------
        :class:`.Atom`
            The ``[atom]`` atom.

        """

        return self._atom

    def clone(self):
        clone = super().clone()
        clone._sulfur = self._sulfur
        clone._hydrogen = self._hydrogen
        clone._atom = self._atom
        return clone

    def with_atoms(self, atom_map):
        clone = super().with_atoms(atom_map)
        clone._sulfur = atom_map.get(
            self._sulfur.get_id(),
            self._sulfur,
        )
        clone._hydrogen = atom_map.get(
            self._hydrogen.get_id(),
            self._hydrogen,
        )
        clone._atom = atom_map.get(
            self._atom.get_id(),
            self._atom,
        )
        return clone

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{self._sulfur}, {self._hydrogen}, {self._atom}, "
            f"bonders={self._bonders}, deleters={self._deleters})"
        )



"""
PDB Writing Utilities
=====================

"""


def _write_pdb_file(self, path, atom_ids):
    """
    Write to a ``.pdb`` file.

    This function should not be used directly, only via
    :meth:`write`.

    Parameters
    ----------
    path : :class:`str`
        The full path to the file being written.

    atom_ids : :class:`iterable` of :class:`int`
        The atom ids of atoms to write. Can be a single
        :class:`int`, if a single atom is to be used, or ``None``,
        if all atoms are to be used.

    Returns
    -------
    None : :class:`NoneType`

    """

    if atom_ids is None:
        atom_ids = range(len(self._atoms))
    elif isinstance(atom_ids, int):
        atom_ids = (atom_ids,)

    lines = []
    atom_counts = {}
    hetatm = "HETATM"
    alt_loc = ""
    res_name = "UNL"
    chain_id = ""
    res_seq = "1"
    i_code = ""
    occupancy = "1.00"
    temp_factor = "0.00"

    coords = self._position_matrix
    # This set will be used by bonds.
    atoms = set()
    for atom in atom_ids:
        atoms.add(atom)

        serial = atom + 1
        element = self._atoms[atom].__class__.__name__
        atom_counts[element] = atom_counts.get(element, 0) + 1
        name = f"{element}{atom_counts[element]}"
        # Make sure the coords are no more than 8 columns wide
        # each.
        x, y, z = (i for i in coords[:, atom])
        lines.append(
            f"{hetatm:<6}{serial:>5} {name:<4}"
            f"{alt_loc:<1}{res_name:<3} {chain_id:<1}"
            f"{res_seq:>4}{i_code:<1}   "
            f" {x:>7.3f} {y:>7.3f} {z:>7.3f}"
            f"{occupancy:>6}{temp_factor:>6}          "
            f"{element:>2}{self._atoms[atom].get_charge():>2}\n"
        )

    conect = "CONECT"
    for bond in self._bonds:
        a1 = bond.get_atom1().get_id()
        a2 = bond.get_atom2().get_id()
        if a1 in atoms and a2 in atoms:
            lines.append(f"{conect:<6}{a1 + 1:>5}{a2 + 1:>5}               \n")

    lines.append("END\n")
    with open(path, "w") as f:
        f.write("".join(lines))



import rdkit.Chem.AllChem as rdkit


def test_get_num_bonds(case_data):
    """
    Test :meth:`.Molecule.get_num_bonds`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and its correct
        SMILES.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_num_bonds(case_data.molecule, case_data.smiles)


def _test_get_num_bonds(molecule, smiles):
    """
    Test :meth:`.Molecule.get_num_bonds`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    smiles : :class:`str`
        The correct SMILES of `molecule`.

    """

    expected = rdkit.MolFromSmiles(smiles, sanitize=False)
    assert molecule.get_num_bonds() == expected.GetNumBonds()



import numpy as np

from ..utilities import is_clone


def test_with_position_matrix(molecule, get_position_matrix):
    """
    Test :meth:`.Molecule.with_position_matrix`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    get_position_matrix : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        position matrix to use for this test.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Save a copy of the position matrix, to ensure the original
    # molecule is not modified by the test, because it is meant to be
    # immutable.
    position_matrix = molecule.get_position_matrix()
    _test_with_position_matrix(molecule, get_position_matrix)
    assert np.all(
        np.equal(
            position_matrix,
            molecule.get_position_matrix(),
        )
    )


def _test_with_position_matrix(molecule, get_position_matrix):
    """
    Test :meth:`.Molecule.with_position_matrix`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    get_position_matrix : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        position matrix to use for this test.

    Returns
    -------
    None : :class:`NoneType`

    """

    position_matrix = get_position_matrix(molecule)
    new = molecule.with_position_matrix(position_matrix)
    is_clone(new, molecule)
    assert np.all(
        np.equal(
            position_matrix,
            new.get_position_matrix(),
        )
    )



import pytest
from pytest_lazyfixture import lazy_fixture

from .hexagonal import *  # noqa

# Fixtures need to be visible for lazy_fixture() calls.
from .honeycomb import *  # noqa
from .kagome import *  # noqa
from .linkerless_honeycomb import *  # noqa
from .periodic_hexagonal import *  # noqa
from .periodic_honeycomb import *  # noqa
from .periodic_kagome import *  # noqa
from .periodic_linkerless_honeycomb import *  # noqa
from .periodic_square import *  # noqa
from .square import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("cof_honeycomb"),
        lazy_fixture("cof_kagome"),
        lazy_fixture("cof_square"),
        lazy_fixture("cof_hexagonal"),
        lazy_fixture("cof_linkerless_honeycomb"),
        lazy_fixture("cof_periodic_honeycomb"),
        lazy_fixture("cof_periodic_kagome"),
        lazy_fixture("cof_periodic_square"),
        lazy_fixture("cof_periodic_hexagonal"),
        lazy_fixture("cof_periodic_linkerless_honeycomb"),
    ),
)
def cof(request):
    return request.param
